#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING
from unittest import mock

import boto3
import pytest
from botocore.exceptions import ClientError
from moto import mock_aws

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook
from airflow.providers.amazon.aws.hooks.glue import GlueDataQualityHook, GlueJobHook
from airflow.providers.amazon.aws.hooks.logs import AwsLogsHook

if TYPE_CHECKING:
    from unittest.mock import MagicMock


class TestGlueJobHook:
    def setup_method(self):
        self.some_aws_region = "us-west-2"

    @mock_aws
    @pytest.mark.parametrize("role_path", ["/", "/custom-path/"])
    def test_get_iam_execution_role(self, role_path):
        expected_role = "my_test_role"
        boto3.client("iam").create_role(
            Path=role_path,
            RoleName=expected_role,
            AssumeRolePolicyDocument=json.dumps(
                {
                    "Version": "2012-10-17",
                    "Statement": {
                        "Effect": "Allow",
                        "Principal": {"Service": "glue.amazonaws.com"},
                        "Action": "sts:AssumeRole",
                    },
                }
            ),
        )

        hook = GlueJobHook(
            aws_conn_id=None,
            job_name="aws_test_glue_job",
            s3_bucket="some_bucket",
            iam_role_name=expected_role,
        )
        iam_role = hook.get_iam_execution_role()
        assert iam_role is not None
        assert "Role" in iam_role
        assert "Arn" in iam_role["Role"]
        assert iam_role["Role"]["Arn"] == f"arn:aws:iam::123456789012:role{role_path}{expected_role}"

    @mock.patch.object(GlueJobHook, "get_iam_execution_role")
    @mock.patch.object(GlueJobHook, "conn")
    def test_init_iam_role_value_error(self, mock_conn, mock_get_iam_execution_role):
        mock_get_iam_execution_role.return_value = mock.MagicMock(
            Role={"RoleName": "my_test_role_name", "RoleArn": "my_test_role"}
        )

        with pytest.raises(ValueError, match="Cannot set iam_role_arn and iam_role_name simultaneously"):
            GlueJobHook(
                job_name="aws_test_glue_job",
                desc="This is test case job from Airflow",
                s3_bucket="some-bucket",
                iam_role_name="my_test_role_name",
                iam_role_arn="my_test_role",
            )

    @mock.patch.object(AwsBaseHook, "conn")
    def test_has_job_exists(self, mock_conn):
        job_name = "aws_test_glue_job"
        mock_conn.get_job.return_value = {"Job": {"Name": job_name}}

        hook = GlueJobHook(aws_conn_id=None, job_name=job_name, s3_bucket="some_bucket")
        result = hook.has_job(job_name)
        assert result is True
        mock_conn.get_job.assert_called_once_with(JobName=hook.job_name)

    @mock.patch.object(AwsBaseHook, "conn")
    def test_has_job_job_doesnt_exists(self, mock_conn):
        class JobNotFoundException(Exception):
            pass

        mock_conn.exceptions.EntityNotFoundException = JobNotFoundException
        mock_conn.get_job.side_effect = JobNotFoundException()

        job_name = "aws_test_glue_job"
        hook = GlueJobHook(aws_conn_id=None, job_name=job_name, s3_bucket="some_bucket")
        result = hook.has_job(job_name)
        assert result is False
        mock_conn.get_job.assert_called_once_with(JobName=job_name)

    @mock.patch.object(GlueJobHook, "get_iam_execution_role")
    @mock.patch.object(AwsBaseHook, "conn")
    def test_role_arn_has_job_exists(self, mock_conn, mock_get_iam_execution_role):
        """
        Calls 'create_or_update_glue_job' with no existing job.
        Should create a new job.
        """

        class JobNotFoundException(Exception):
            pass

        expected_job_name = "aws_test_glue_job"
        job_description = "This is test case job from Airflow"
        role_name = "my_test_role"
        role_name_arn = "test_role"
        some_s3_bucket = "bucket"

        mock_conn.exceptions.EntityNotFoundException = JobNotFoundException
        mock_conn.get_job.side_effect = JobNotFoundException()
        mock_get_iam_execution_role.return_value = {"Role": {"RoleName": role_name, "Arn": role_name_arn}}

        hook = GlueJobHook(
            s3_bucket=some_s3_bucket,
            job_name=expected_job_name,
            desc=job_description,
            concurrent_run_limit=2,
            retry_limit=3,
            num_of_dpus=5,
            iam_role_arn=role_name_arn,
            create_job_kwargs={"Command": {}},
            region_name=self.some_aws_region,
            update_config=True,
        )

        result = hook.create_or_update_glue_job()

        mock_conn.get_job.assert_called_once_with(JobName=expected_job_name)
        mock_conn.create_job.assert_called_once_with(
            Command={},
            Description=job_description,
            ExecutionProperty={"MaxConcurrentRuns": 2},
            LogUri=f"s3://{some_s3_bucket}/logs/glue-logs/{expected_job_name}",
            MaxCapacity=5,
            MaxRetries=3,
            Name=expected_job_name,
            Role=role_name_arn,
        )
        mock_conn.update_job.assert_not_called()
        assert result == expected_job_name

    @mock.patch.object(GlueJobHook, "get_iam_execution_role")
    @mock.patch.object(GlueJobHook, "conn")
    def test_create_or_update_glue_job_create_new_job(self, mock_conn, mock_get_iam_execution_role):
        """
        Calls 'create_or_update_glue_job' with no existing job.
        Should create a new job.
        """

        class JobNotFoundException(Exception):
            pass

        expected_job_name = "aws_test_glue_job"
        job_description = "This is test case job from Airflow"
        role_name = "my_test_role"
        role_name_arn = "test_role"
        some_s3_bucket = "bucket"

        mock_conn.exceptions.EntityNotFoundException = JobNotFoundException
        mock_conn.get_job.side_effect = JobNotFoundException()
        mock_get_iam_execution_role.return_value = {"Role": {"RoleName": role_name, "Arn": role_name_arn}}

        hook = GlueJobHook(
            s3_bucket=some_s3_bucket,
            job_name=expected_job_name,
            desc=job_description,
            concurrent_run_limit=2,
            retry_limit=3,
            num_of_dpus=5,
            iam_role_name=role_name,
            create_job_kwargs={"Command": {}},
            region_name=self.some_aws_region,
            update_config=True,
        )

        result = hook.create_or_update_glue_job()

        mock_conn.get_job.assert_called_once_with(JobName=expected_job_name)
        mock_conn.create_job.assert_called_once_with(
            Command={},
            Description=job_description,
            ExecutionProperty={"MaxConcurrentRuns": 2},
            LogUri=f"s3://{some_s3_bucket}/logs/glue-logs/{expected_job_name}",
            MaxCapacity=5,
            MaxRetries=3,
            Name=expected_job_name,
            Role=role_name_arn,
        )
        mock_conn.update_job.assert_not_called()
        assert result == expected_job_name

    @mock.patch.object(GlueJobHook, "get_iam_execution_role")
    @mock.patch.object(GlueJobHook, "conn")
    def test_create_or_update_glue_job_create_new_job_without_s3_bucket(
        self, mock_conn, mock_get_iam_execution_role
    ):
        """
        Calls 'create_or_update_glue_job' with no existing job.
        Should create a new job.
        """

        class JobNotFoundException(Exception):
            pass

        expected_job_name = "aws_test_glue_job"
        job_description = "This is test case job from Airflow"
        role_name = "my_test_role"
        role_name_arn = "test_role"

        mock_conn.exceptions.EntityNotFoundException = JobNotFoundException
        mock_conn.get_job.side_effect = JobNotFoundException()
        mock_get_iam_execution_role.return_value = {"Role": {"RoleName": role_name, "Arn": role_name_arn}}

        hook = GlueJobHook(
            job_name=expected_job_name,
            desc=job_description,
            concurrent_run_limit=2,
            retry_limit=3,
            num_of_dpus=5,
            iam_role_name=role_name,
            create_job_kwargs={"Command": {}},
            region_name=self.some_aws_region,
            update_config=True,
        )

        result = hook.create_or_update_glue_job()

        mock_conn.get_job.assert_called_once_with(JobName=expected_job_name)
        mock_conn.create_job.assert_called_once_with(
            Command={},
            Description=job_description,
            ExecutionProperty={"MaxConcurrentRuns": 2},
            MaxCapacity=5,
            MaxRetries=3,
            Name=expected_job_name,
            Role=role_name_arn,
        )
        mock_conn.update_job.assert_not_called()
        assert result == expected_job_name

    @mock.patch.object(GlueJobHook, "get_iam_execution_role")
    @mock.patch.object(GlueJobHook, "conn")
    def test_create_or_update_glue_job_update_existing_job(self, mock_conn, mock_get_iam_execution_role):
        """
        Calls 'create_or_update_glue_job' with a existing job.
        Should update existing job configurations.
        """
        job_name = "aws_test_glue_job"
        job_description = "This is test case job from Airflow"
        role_name = "my_test_role"
        role_name_arn = "test_role"
        some_script = "s3://glue-examples/glue-scripts/sample_aws_glue_job.py"
        some_s3_bucket = "my-includes"

        mock_conn.get_job.return_value = {
            "Job": {
                "Name": job_name,
                "Description": "Old description of job",
                "Role": role_name_arn,
            }
        }
        mock_get_iam_execution_role.return_value = {"Role": {"RoleName": role_name, "Arn": "test_role"}}

        hook = GlueJobHook(
            job_name=job_name,
            desc=job_description,
            script_location=some_script,
            iam_role_name=role_name,
            s3_bucket=some_s3_bucket,
            region_name=self.some_aws_region,
            update_config=True,
        )

        result = hook.create_or_update_glue_job()

        assert mock_conn.get_job.call_count == 2
        mock_conn.update_job.assert_called_once_with(
            JobName=job_name,
            JobUpdate={
                "Description": job_description,
                "LogUri": f"s3://{some_s3_bucket}/logs/glue-logs/{job_name}",
                "Role": role_name_arn,
                "ExecutionProperty": {"MaxConcurrentRuns": 1},
                "Command": {"Name": "glueetl", "ScriptLocation": some_script},
                "MaxRetries": 0,
                "MaxCapacity": 10,
            },
        )
        assert result == job_name

    @mock_aws
    @mock.patch.object(GlueJobHook, "get_iam_execution_role")
    def test_create_or_update_glue_job_worker_type(self, mock_get_iam_execution_role):
        mock_get_iam_execution_role.return_value = {"Role": {"RoleName": "my_test_role", "Arn": "test_role"}}
        some_script = "s3:/glue-examples/glue-scripts/sample_aws_glue_job.py"
        some_s3_bucket = "my-includes"
        expected_job_name = "aws_test_glue_job_worker_type"

        glue_job = GlueJobHook(
            job_name=expected_job_name,
            desc="This is test case job from Airflow",
            script_location=some_script,
            iam_role_name="my_test_role",
            s3_bucket=some_s3_bucket,
            region_name=self.some_aws_region,
            create_job_kwargs={"WorkerType": "G.2X", "NumberOfWorkers": 60},
            update_config=True,
        )

        result = glue_job.create_or_update_glue_job()

        assert result == expected_job_name

    @mock.patch.object(GlueJobHook, "get_iam_execution_role")
    @mock.patch.object(GlueJobHook, "conn")
    def test_init_worker_type_value_error(self, mock_conn, mock_get_iam_execution_role):
        mock_get_iam_execution_role.return_value = mock.MagicMock(Role={"RoleName": "my_test_role"})
        some_script = "s3:/glue-examples/glue-scripts/sample_aws_glue_job.py"
        some_s3_bucket = "my-includes"

        with pytest.raises(ValueError, match="Cannot specify num_of_dpus with custom WorkerType"):
            GlueJobHook(
                job_name="aws_test_glue_job",
                desc="This is test case job from Airflow",
                script_location=some_script,
                iam_role_name="my_test_role",
                s3_bucket=some_s3_bucket,
                region_name=self.some_aws_region,
                num_of_dpus=20,
                create_job_kwargs={"WorkerType": "G.2X", "NumberOfWorkers": 60},
                update_config=True,
            )

    @mock.patch.object(GlueJobHook, "get_job_state")
    @mock.patch.object(GlueJobHook, "conn")
    def test_initialize_job(self, mock_conn, mock_get_job_state):
        some_data_path = "s3://glue-datasets/examples/medicare/SampleData.csv"
        some_script_arguments = {"--s3_input_data_path": some_data_path}
        some_run_kwargs = {"NumberOfWorkers": 5}
        some_script = "s3:/glue-examples/glue-scripts/sample_aws_glue_job.py"
        some_s3_bucket = "my-includes"

        mock_conn.start_job_run()

        mock_job_run_state = mock_get_job_state.return_value
        glue_job_hook = GlueJobHook(
            job_name="aws_test_glue_job",
            desc="This is test case job from Airflow",
            iam_role_name="my_test_role",
            script_location=some_script,
            s3_bucket=some_s3_bucket,
            region_name=self.some_aws_region,
            update_config=False,
        )
        glue_job_run = glue_job_hook.initialize_job(some_script_arguments, some_run_kwargs)
        glue_job_run_state = glue_job_hook.get_job_state(glue_job_run["JobName"], glue_job_run["JobRunId"])
        assert glue_job_run_state == mock_job_run_state, "Mocks but be equal"

    @mock.patch.object(AwsLogsHook, "get_conn")
    @mock.patch.object(GlueJobHook, "conn")
    def test_print_job_logs_returns_token(self, conn_mock: MagicMock, log_client_mock: MagicMock, caplog):
        hook = GlueJobHook(job_name="test")
        conn_mock().get_job_run.return_value = {"JobRun": {"LogGroupName": "my_log_group"}}
        log_client_mock().get_paginator().paginate.return_value = [
            # first response : 2 log lines
            {
                "events": [
                    {"logStreamName": "stream", "timestamp": 123, "message": "hello\n"},
                    {"logStreamName": "stream", "timestamp": 123, "message": "world\n"},
                ],
                "searchedLogStreams": [],
                "nextToken": "my_continuation_token",
                "ResponseMetadata": {"HTTPStatusCode": 200},
            },
            # second response, reached end of stream
            {"events": [], "searchedLogStreams": [], "ResponseMetadata": {"HTTPStatusCode": 200}},
        ]

        tokens = GlueJobHook.LogContinuationTokens()
        with caplog.at_level("INFO"):
            hook.print_job_logs("name", "run", tokens)
        assert "\thello\n\tworld\n" in caplog.text
        assert tokens.output_stream_continuation == "my_continuation_token"
        assert tokens.error_stream_continuation == "my_continuation_token"

    @mock.patch.object(AwsLogsHook, "get_conn")
    @mock.patch.object(GlueJobHook, "conn")
    def test_print_job_logs_no_stream_yet(self, conn_mock: MagicMock, client_mock: MagicMock):
        hook = GlueJobHook()
        conn_mock().get_job_run.return_value = {"JobRun": {"LogGroupName": "my_log_group"}}
        client_mock().get_paginator().paginate.side_effect = ClientError(
            {"Error": {"Code": "ResourceNotFoundException"}}, "op"
        )

        tokens = GlueJobHook.LogContinuationTokens()
        hook.print_job_logs("name", "run", tokens)  # should not error

        assert tokens.output_stream_continuation is None
        assert tokens.error_stream_continuation is None
        assert client_mock().get_paginator().paginate.call_count == 2

    @mock.patch.object(GlueJobHook, "get_job_state")
    def test_job_completion_success(self, get_state_mock: MagicMock):
        hook = GlueJobHook(job_poll_interval=0)
        get_state_mock.side_effect = [
            "RUNNING",
            "RUNNING",
            "SUCCEEDED",
        ]

        hook.job_completion("job_name", "run_id")

        assert get_state_mock.call_count == 3
        get_state_mock.assert_called_with("job_name", "run_id")

    @mock.patch.object(GlueJobHook, "get_job_state")
    def test_job_completion_failure(self, get_state_mock: MagicMock):
        hook = GlueJobHook(job_poll_interval=0)
        get_state_mock.side_effect = [
            "RUNNING",
            "RUNNING",
            "FAILED",
        ]

        with pytest.raises(AirflowException):
            hook.job_completion("job_name", "run_id")

        assert get_state_mock.call_count == 3

    @pytest.mark.asyncio
    @mock.patch.object(GlueJobHook, "async_get_job_state")
    async def test_async_job_completion_success(self, get_state_mock: MagicMock):
        hook = GlueJobHook(job_poll_interval=0)
        get_state_mock.side_effect = [
            "RUNNING",
            "RUNNING",
            "SUCCEEDED",
        ]

        await hook.async_job_completion("job_name", "run_id")

        assert get_state_mock.call_count == 3
        get_state_mock.assert_called_with("job_name", "run_id")

    @pytest.mark.asyncio
    @mock.patch.object(GlueJobHook, "async_get_job_state")
    async def test_async_job_completion_failure(self, get_state_mock: MagicMock):
        hook = GlueJobHook(job_poll_interval=0)
        get_state_mock.side_effect = [
            "RUNNING",
            "RUNNING",
            "FAILED",
        ]

        with pytest.raises(AirflowException):
            await hook.async_job_completion("job_name", "run_id")

        assert get_state_mock.call_count == 3

    @mock.patch.object(GlueJobHook, "conn")
    def test_get_job_state_success(self, mock_conn):
        hook = GlueJobHook()
        job_name = "test_job"
        run_id = "test_run_id"
        expected_state = "SUCCEEDED"

        mock_conn.get_job_run.return_value = {"JobRun": {"JobRunState": expected_state}}

        result = hook.get_job_state(job_name, run_id)

        assert result == expected_state
        mock_conn.get_job_run.assert_called_once_with(
            JobName=job_name, RunId=run_id, PredecessorsIncluded=True
        )

    @mock.patch.object(GlueJobHook, "conn")
    def test_get_job_state_retry_on_client_error(self, mock_conn):
        hook = GlueJobHook()
        job_name = "test_job"
        run_id = "test_run_id"
        expected_state = "SUCCEEDED"

        mock_conn.get_job_run.side_effect = [
            ClientError(
                {"Error": {"Code": "ThrottlingException", "Message": "Rate exceeded"}}, "get_job_run"
            ),
            {"JobRun": {"JobRunState": expected_state}},
        ]

        result = hook.get_job_state(job_name, run_id)

        assert result == expected_state
        assert mock_conn.get_job_run.call_count == 2

    @mock.patch.object(GlueJobHook, "conn")
    def test_get_job_state_fails_after_all_retries(self, mock_conn):
        """Test get_job_state raises exception when all retries are exhausted."""
        hook = GlueJobHook()
        job_name = "test_job"
        run_id = "test_run_id"

        mock_conn.get_job_run.side_effect = ClientError(
            {"Error": {"Code": "ThrottlingException", "Message": "Rate exceeded"}}, "get_job_run"
        )

        with pytest.raises(ClientError) as exc_info:
            hook.get_job_state(job_name, run_id)

        assert exc_info.value.response["Error"]["Code"] == "ThrottlingException"
        assert mock_conn.get_job_run.call_count == 5

    @pytest.mark.asyncio
    @mock.patch.object(GlueJobHook, "get_async_conn")
    async def test_async_get_job_state_success(self, mock_get_async_conn):
        hook = GlueJobHook()
        job_name = "test_job"
        run_id = "test_run_id"
        expected_state = "RUNNING"

        mock_client = mock.AsyncMock()
        mock_client.get_job_run.return_value = {"JobRun": {"JobRunState": expected_state}}
        mock_context = mock.AsyncMock()
        mock_context.__aenter__.return_value = mock_client
        mock_context.__aexit__.return_value = None
        mock_get_async_conn.return_value = mock_context

        result = await hook.async_get_job_state(job_name, run_id)

        assert result == expected_state
        mock_client.get_job_run.assert_called_once_with(JobName=job_name, RunId=run_id)

    @pytest.mark.asyncio
    @mock.patch.object(GlueJobHook, "get_async_conn")
    async def test_async_get_job_state_retry_on_client_error(self, mock_get_async_conn):
        hook = GlueJobHook()
        job_name = "test_job"
        run_id = "test_run_id"
        expected_state = "FAILED"

        mock_client = mock.AsyncMock()
        mock_client.get_job_run.side_effect = [
            ClientError(
                {"Error": {"Code": "ServiceUnavailable", "Message": "Service temporarily unavailable"}},
                "get_job_run",
            ),
            {"JobRun": {"JobRunState": expected_state}},
        ]
        mock_context = mock.AsyncMock()
        mock_context.__aenter__.return_value = mock_client
        mock_context.__aexit__.return_value = None
        mock_get_async_conn.return_value = mock_context

        result = await hook.async_get_job_state(job_name, run_id)

        assert result == expected_state
        assert mock_client.get_job_run.call_count == 2

    @pytest.mark.asyncio
    @mock.patch.object(GlueJobHook, "get_async_conn")
    async def test_async_get_job_state_fails_after_all_retries(self, mock_get_async_conn):
        hook = GlueJobHook()
        job_name = "test_job"
        run_id = "test_run_id"

        mock_client = mock.AsyncMock()
        mock_client.get_job_run.side_effect = ClientError(
            {"Error": {"Code": "ServiceUnavailable", "Message": "Service temporarily unavailable"}},
            "get_job_run",
        )
        mock_context = mock.AsyncMock()
        mock_context.__aenter__.return_value = mock_client
        mock_context.__aexit__.return_value = None
        mock_get_async_conn.return_value = mock_context

        with pytest.raises(ClientError) as exc_info:
            await hook.async_get_job_state(job_name, run_id)

        assert exc_info.value.response["Error"]["Code"] == "ServiceUnavailable"
        assert mock_client.get_job_run.call_count == 5


class TestGlueDataQualityHook:
    RUN_ID = "1234"
    RULE_SET_NAME = "test_rule"
    RULE_SET_CONFIG = {
        "Name": "test_rule",
        "Ruleset": 'Rules=[ColumnLength "review_id" = 15]',
        "TargetTable": {"DatabaseName": "test_db", "TableName": "test_table"},
        "Description": "test rule",
    }

    def setup_method(self):
        self.glue = GlueDataQualityHook()

    def test_glue_data_quality_hook(self):
        glue_data_quality_hook = GlueDataQualityHook()
        assert glue_data_quality_hook.conn is not None
        assert glue_data_quality_hook.aws_conn_id == "aws_default"

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_data_quality_ruleset_exists(self, mock_conn):
        mock_conn.get_data_quality_ruleset.return_value = {"Name": self.RULE_SET_NAME}

        result = self.glue.has_data_quality_ruleset(name=self.RULE_SET_NAME)

        assert result is True
        mock_conn.get_data_quality_ruleset.assert_called_once_with(Name=self.RULE_SET_NAME)

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_quality_ruleset_doesnt_exists(self, mock_conn):
        error_message = f"Cannot find Data Quality Ruleset in account 1234567 with name {self.RULE_SET_NAME}"

        err_response = {"Error": {"Code": "EntityNotFoundException", "Message": error_message}}

        exception = boto3.client("glue").exceptions.ClientError(err_response, "test")
        returned_exception = type(exception)

        mock_conn.exceptions.EntityNotFoundException = returned_exception
        mock_conn.get_data_quality_ruleset.side_effect = exception

        result = self.glue.has_data_quality_ruleset(name=self.RULE_SET_NAME)

        assert result is False
        mock_conn.get_data_quality_ruleset.assert_called_once_with(Name=self.RULE_SET_NAME)

    @mock.patch.object(AwsBaseHook, "conn")
    def test_validate_evaluation_results(self, mock_conn, caplog):
        response_evaluation_run = {"RunId": self.RUN_ID, "ResultIds": ["resultId1"]}

        response_batch_result = {
            "RunId": self.RUN_ID,
            "ResultIds": ["resultId1"],
            "Results": [
                {
                    "ResultId": "resultId1",
                    "RulesetName": "rulesetOne",
                    "RuleResults": [
                        {
                            "Name": "Rule_1",
                            "Description": "RowCount between 150000 and 600000",
                            "EvaluatedMetrics": {"Dataset.*.RowCount": 300000.0},
                            "Result": "PASS",
                        }
                    ],
                }
            ],
        }
        mock_conn.get_data_quality_ruleset_evaluation_run.return_value = response_evaluation_run

        mock_conn.batch_get_data_quality_result.return_value = response_batch_result

        with caplog.at_level(logging.INFO, logger=self.glue.log.name):
            caplog.clear()
            self.glue.validate_evaluation_run_results(evaluation_run_id=self.RUN_ID, show_results=False)

        mock_conn.get_data_quality_ruleset_evaluation_run.assert_called_once_with(RunId=self.RUN_ID)
        mock_conn.batch_get_data_quality_result.assert_called_once_with(
            ResultIds=response_evaluation_run["ResultIds"]
        )

        assert caplog.messages == [
            "AWS Glue data quality ruleset evaluation run, total number of rules failed: 0"
        ]

    @mock.patch.object(AwsBaseHook, "conn")
    def test_validate_evaluation_results_should_fail_when_any_rules_failed(self, mock_conn, caplog):
        response_batch_result = {
            "RunId": self.RUN_ID,
            "ResultIds": ["resultId1"],
            "Results": [
                {
                    "ResultId": "resultId1",
                    "RulesetName": "rulesetOne",
                    "RuleResults": [
                        {
                            "Name": "Rule_1",
                            "Description": "RowCount between 150000 and 600000",
                            "EvaluatedMetrics": {"Dataset.*.RowCount": 300000.0},
                            "Result": "PASS",
                        },
                        {
                            "Name": "Rule_2",
                            "Description": "ColumnLength 'marketplace' between 1 and 2",
                            "EvaluationMessage": "Value: 9.0 does not meet the constraint requirement!",
                            "Result": "FAIL",
                            "EvaluatedMetrics": {
                                "Column.marketplace.MaximumLength": 9.0,
                                "Column.marketplace.MinimumLength": 2.0,
                            },
                        },
                    ],
                }
            ],
        }

        response_evaluation_run = {"RunId": self.RUN_ID, "ResultIds": ["resultId1"]}

        mock_conn.get_data_quality_ruleset_evaluation_run.return_value = response_evaluation_run

        mock_conn.batch_get_data_quality_result.return_value = response_batch_result

        with caplog.at_level(logging.INFO, logger=self.glue.log.name):
            caplog.clear()

            with pytest.raises(
                AirflowException,
                match="AWS Glue data quality ruleset evaluation run failed for one or more rules",
            ):
                self.glue.validate_evaluation_run_results(evaluation_run_id=self.RUN_ID, show_results=False)

            mock_conn.get_data_quality_ruleset_evaluation_run.assert_called_once_with(RunId=self.RUN_ID)
            mock_conn.batch_get_data_quality_result.assert_called_once_with(
                ResultIds=response_evaluation_run["ResultIds"]
            )

        assert caplog.messages == [
            "AWS Glue data quality ruleset evaluation run, total number of rules failed: 1"
        ]

    @mock.patch.object(GlueDataQualityHook, "conn")
    def test_log_recommendation_results(self, glue_data_quality_hook_mock_conn, caplog):
        rules = """ Rules = [
                                RowCount between 2 and 8,
                                IsComplete "name"
                                ]
                    """
        glue_data_quality_hook_mock_conn.get_data_quality_rule_recommendation_run.return_value = {
            "RunId": self.RUN_ID,
            "DataSource": {"GlueTable": {"DatabaseName": "TestDB", "TableName": "TestTable"}},
            "RecommendedRuleset": rules,
        }

        with caplog.at_level(logging.INFO, logger=self.glue.log.name):
            self.glue.log_recommendation_results(run_id=self.RUN_ID)

        glue_data_quality_hook_mock_conn.get_data_quality_rule_recommendation_run.assert_called_once_with(
            RunId=self.RUN_ID
        )
        assert rules in caplog.messages



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING

from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.version_compat import BaseOperator
from airflow.providers.ftp.hooks.ftp import FTPHook

if TYPE_CHECKING:
    from airflow.utils.context import Context


class FTPToS3Operator(BaseOperator):
    """
    Transfer of one or more files from an FTP server to S3.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:FTPToS3Operator`

    :param ftp_path: The ftp remote path. For one file it is mandatory to include the file as well.
        For multiple files, it is the route where the files will be found.
    :param s3_bucket: The targeted s3 bucket in which to upload the file(s).
    :param s3_key: The targeted s3 key. For one file it must include the file path. For several,
        it must end with "/".
    :param ftp_filenames: Only used if you want to move multiple files. You can pass a list
        with exact filenames present in the ftp path, or a prefix that all files must meet. It
        can also be the string '*' for moving all the files within the ftp path.
    :param s3_filenames: Only used if you want to move multiple files and name them different from
        the originals from the ftp. It can be a list of filenames or file prefix (that will replace
        the ftp prefix).
    :param ftp_conn_id: The ftp connection id. The name or identifier for
        establishing a connection to the FTP server.
    :param aws_conn_id: The s3 connection id. The name or identifier for
        establishing a connection to S3.
    :param replace: A flag to decide whether or not to overwrite the key
        if it already exists. If replace is False and the key exists, an
        error will be raised.
    :param encrypt: If True, the file will be encrypted on the server-side
        by S3 and will be stored in an encrypted form while at rest in S3.
    :param gzip: If True, the file will be compressed locally
    :param acl_policy: String specifying the canned ACL policy for the file being
        uploaded to the S3 bucket.
    """

    template_fields: Sequence[str] = ("ftp_path", "s3_bucket", "s3_key", "ftp_filenames", "s3_filenames")

    def __init__(
        self,
        *,
        ftp_path: str,
        s3_bucket: str,
        s3_key: str,
        ftp_filenames: str | list[str] | None = None,
        s3_filenames: str | list[str] | None = None,
        ftp_conn_id: str = "ftp_default",
        aws_conn_id: str | None = "aws_default",
        replace: bool = False,
        encrypt: bool = False,
        gzip: bool = False,
        acl_policy: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.ftp_path = ftp_path
        self.s3_bucket = s3_bucket
        self.s3_key = s3_key
        self.ftp_filenames = ftp_filenames
        self.s3_filenames = s3_filenames
        self.aws_conn_id = aws_conn_id
        self.ftp_conn_id = ftp_conn_id
        self.replace = replace
        self.encrypt = encrypt
        self.gzip = gzip
        self.acl_policy = acl_policy
        self.s3_hook: S3Hook | None = None
        self.ftp_hook: FTPHook | None = None

    def __upload_to_s3_from_ftp(self, remote_filename, s3_file_key):
        with NamedTemporaryFile() as local_tmp_file:
            self.ftp_hook.retrieve_file(
                remote_full_path=remote_filename, local_full_path_or_buffer=local_tmp_file.name
            )

            self.s3_hook.load_file(
                filename=local_tmp_file.name,
                key=s3_file_key,
                bucket_name=self.s3_bucket,
                replace=self.replace,
                encrypt=self.encrypt,
                gzip=self.gzip,
                acl_policy=self.acl_policy,
            )
            self.log.info("File upload to %s", s3_file_key)

    def execute(self, context: Context):
        self.ftp_hook = FTPHook(ftp_conn_id=self.ftp_conn_id)
        self.s3_hook = S3Hook(self.aws_conn_id)

        if self.ftp_filenames:
            if isinstance(self.ftp_filenames, str):
                self.log.info("Getting files in %s", self.ftp_path)

                list_dir = self.ftp_hook.list_directory(
                    path=self.ftp_path,
                )

                if self.ftp_filenames == "*":
                    files = list_dir
                else:
                    ftp_filename: str = self.ftp_filenames
                    files = [f for f in list_dir if ftp_filename in f]

                for file in files:
                    self.log.info("Moving file %s", file)

                    if self.s3_filenames and isinstance(self.s3_filenames, str):
                        filename = file.replace(self.ftp_filenames, self.s3_filenames)
                    else:
                        filename = file

                    s3_file_key = f"{self.s3_key}{filename}"
                    self.__upload_to_s3_from_ftp(file, s3_file_key)

            else:
                if self.s3_filenames:
                    for ftp_file, s3_file in zip(self.ftp_filenames, self.s3_filenames):
                        self.__upload_to_s3_from_ftp(self.ftp_path + ftp_file, self.s3_key + s3_file)
                else:
                    for ftp_file in self.ftp_filenames:
                        self.__upload_to_s3_from_ftp(self.ftp_path + ftp_file, self.s3_key + ftp_file)
        else:
            self.__upload_to_s3_from_ftp(self.ftp_path, self.s3_key)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains sensors for AWS CloudFormation."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields

if TYPE_CHECKING:
    from airflow.utils.context import Context

from airflow.providers.amazon.aws.hooks.cloud_formation import CloudFormationHook


class CloudFormationCreateStackSensor(AwsBaseSensor[CloudFormationHook]):
    """
    Waits for a stack to be created successfully on AWS CloudFormation.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:CloudFormationCreateStackSensor`

    :param stack_name: The name of the stack to wait for (templated)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = CloudFormationHook
    template_fields: Sequence[str] = aws_template_fields("stack_name")
    ui_color = "#C5CAE9"

    def __init__(self, *, stack_name, **kwargs):
        super().__init__(**kwargs)
        self.stack_name = stack_name

    def poke(self, context: Context):
        stack_status = self.hook.get_stack_status(self.stack_name)
        if stack_status == "CREATE_COMPLETE":
            return True
        if stack_status in ("CREATE_IN_PROGRESS", None):
            return False

        raise ValueError(f"Stack {self.stack_name} in bad state: {stack_status}")


class CloudFormationDeleteStackSensor(AwsBaseSensor[CloudFormationHook]):
    """
    Waits for a stack to be deleted successfully on AWS CloudFormation.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:CloudFormationDeleteStackSensor`

    :param stack_name: The name of the stack to wait for (templated)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = CloudFormationHook
    template_fields: Sequence[str] = aws_template_fields("stack_name")
    ui_color = "#C5CAE9"

    def __init__(
        self,
        *,
        stack_name: str,
        aws_conn_id: str | None = "aws_default",
        region_name: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.aws_conn_id = aws_conn_id
        self.region_name = region_name
        self.stack_name = stack_name

    def poke(self, context: Context):
        stack_status = self.hook.get_stack_status(self.stack_name)
        if stack_status in ("DELETE_COMPLETE", None):
            return True
        if stack_status == "DELETE_IN_PROGRESS":
            return False

        raise ValueError(f"Stack {self.stack_name} in bad state: {stack_status}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

import boto3

from airflow.decorators import task
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.hooks.ecs import EcsTaskStates
from airflow.providers.amazon.aws.operators.ecs import EcsRunTaskOperator
from airflow.providers.amazon.aws.sensors.ecs import EcsTaskStateSensor
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

DAG_ID = "example_ecs_fargate"

# Externally fetched variables:
SUBNETS_KEY = "SUBNETS"  # At least one public subnet is required.
SECURITY_GROUPS_KEY = "SECURITY_GROUPS"

sys_test_context_task = (
    SystemTestContextBuilder()
    .add_variable(SUBNETS_KEY, split_string=True)
    .add_variable(SECURITY_GROUPS_KEY, split_string=True)
    .build()
)


@task
def create_cluster(cluster_name: str) -> None:
    """Creates an ECS cluster."""
    boto3.client("ecs").create_cluster(clusterName=cluster_name)


@task
def register_task_definition(task_name: str, container_name: str) -> str:
    """Creates a Task Definition."""
    response = boto3.client("ecs").register_task_definition(
        family=task_name,
        # CPU and Memory are required for Fargate and are set to the lowest currently allowed values.
        cpu="256",
        memory="512",
        containerDefinitions=[
            {
                "name": container_name,
                "image": "ubuntu",
                "workingDirectory": "/usr/bin",
                "entryPoint": ["sh", "-c"],
                "command": ["ls"],
            }
        ],
        requiresCompatibilities=["FARGATE"],
        networkMode="awsvpc",
    )

    return response["taskDefinition"]["taskDefinitionArn"]


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_task_definition(task_definition_arn: str) -> None:
    """Deletes the Task Definition."""
    boto3.client("ecs").deregister_task_definition(taskDefinition=task_definition_arn)


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_cluster(cluster_name: str) -> None:
    """Deletes the ECS cluster."""
    boto3.client("ecs").delete_cluster(cluster=cluster_name)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]

    cluster_name = f"{env_id}-test-cluster"
    container_name = f"{env_id}-test-container"
    task_definition_name = f"{env_id}-test-definition"

    create_task_definition = register_task_definition(task_definition_name, container_name)

    # [START howto_operator_ecs]
    hello_world = EcsRunTaskOperator(
        task_id="hello_world",
        cluster=cluster_name,
        task_definition=task_definition_name,
        launch_type="FARGATE",
        overrides={
            "containerOverrides": [
                {
                    "name": container_name,
                    "command": ["echo", "hello", "world"],
                },
            ],
        },
        network_configuration={
            "awsvpcConfiguration": {
                "subnets": test_context[SUBNETS_KEY],
                "securityGroups": test_context[SECURITY_GROUPS_KEY],
                "assignPublicIp": "ENABLED",
            },
        },
    )
    # [END howto_operator_ecs]

    # EcsRunTaskOperator waits by default, setting as False to test the Sensor below.
    hello_world.wait_for_completion = False

    # [START howto_sensor_ecs_task_state]
    # By default, EcsTaskStateSensor waits until the task has started, but the
    # demo task runs so fast that the sensor misses it.  This sensor instead
    # demonstrates how to wait until the ECS Task has completed by providing
    # the target_state and failure_states parameters.
    await_task_finish = EcsTaskStateSensor(
        task_id="await_task_finish",
        cluster=cluster_name,
        task=hello_world.output["ecs_task_arn"],
        target_state=EcsTaskStates.STOPPED,
        failure_states={EcsTaskStates.NONE},
    )
    # [END howto_sensor_ecs_task_state]

    chain(
        # TEST SETUP
        test_context,
        create_cluster(cluster_name),
        create_task_definition,
        # TEST BODY
        hello_world,
        await_task_finish,
        # TEST TEARDOWN
        delete_task_definition(create_task_definition),
        delete_cluster(cluster_name),
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.providers.amazon.aws.hooks.neptune import NeptuneHook
from airflow.providers.amazon.aws.triggers.base import AwsBaseWaiterTrigger

if TYPE_CHECKING:
    from airflow.providers.amazon.aws.hooks.base_aws import AwsGenericHook


class NeptuneClusterAvailableTrigger(AwsBaseWaiterTrigger):
    """
    Triggers when a Neptune Cluster is available.

    :param db_cluster_id: Cluster ID to poll.
    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    :param region_name: AWS region name (example: us-east-1)
    """

    def __init__(
        self,
        *,
        db_cluster_id: str,
        waiter_delay: int = 30,
        waiter_max_attempts: int = 60,
        aws_conn_id: str | None = None,
        region_name: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            serialized_fields={"db_cluster_id": db_cluster_id},
            waiter_name="cluster_available",
            waiter_args={"DBClusterIdentifier": db_cluster_id},
            failure_message="Failed to start Neptune cluster",
            status_message="Status of Neptune cluster is",
            status_queries=["DBClusters[0].Status"],
            return_key="db_cluster_id",
            return_value=db_cluster_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
            **kwargs,
        )

    def hook(self) -> AwsGenericHook:
        return NeptuneHook(
            aws_conn_id=self.aws_conn_id,
            region_name=self.region_name,
            verify=self.verify,
            config=self.botocore_config,
        )


class NeptuneClusterStoppedTrigger(AwsBaseWaiterTrigger):
    """
    Triggers when a Neptune Cluster is stopped.

    :param db_cluster_id: Cluster ID to poll.
    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    :param region_name: AWS region name (example: us-east-1)
    """

    def __init__(
        self,
        *,
        db_cluster_id: str,
        waiter_delay: int = 30,
        waiter_max_attempts: int = 60,
        aws_conn_id: str | None = None,
        region_name: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            serialized_fields={"db_cluster_id": db_cluster_id},
            waiter_name="cluster_stopped",
            waiter_args={"DBClusterIdentifier": db_cluster_id},
            failure_message="Failed to stop Neptune cluster",
            status_message="Status of Neptune cluster is",
            status_queries=["DBClusters[0].Status"],
            return_key="db_cluster_id",
            return_value=db_cluster_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
            **kwargs,
        )

    def hook(self) -> AwsGenericHook:
        return NeptuneHook(
            aws_conn_id=self.aws_conn_id,
            region_name=self.region_name,
            verify=self.verify,
            config=self.botocore_config,
        )


class NeptuneClusterInstancesAvailableTrigger(AwsBaseWaiterTrigger):
    """
    Triggers when a Neptune Cluster Instance is available.

    :param db_cluster_id: Cluster ID to wait on instances from
    :param waiter_delay: The amount of time in seconds to wait between attempts.
    :param waiter_max_attempts: The maximum number of attempts to be made.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    :param region_name: AWS region name (example: us-east-1)
    """

    def __init__(
        self,
        *,
        db_cluster_id: str,
        waiter_delay: int = 30,
        waiter_max_attempts: int = 60,
        aws_conn_id: str | None = None,
        region_name: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            serialized_fields={"db_cluster_id": db_cluster_id},
            waiter_name="db_instance_available",
            waiter_args={"Filters": [{"Name": "db-cluster-id", "Values": [db_cluster_id]}]},
            failure_message="Failed to start Neptune instances",
            status_message="Status of Neptune instances are",
            status_queries=["DBInstances[].Status"],
            return_key="db_cluster_id",
            return_value=db_cluster_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
            **kwargs,
        )

    def hook(self) -> AwsGenericHook:
        return NeptuneHook(
            aws_conn_id=self.aws_conn_id,
            region_name=self.region_name,
            verify=self.verify,
            config=self.botocore_config,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Any
from unittest import mock

import pytest

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.sensors.emr import EmrNotebookExecutionSensor


class TestEmrNotebookExecutionSensor:
    def _generate_response(self, status: str, reason: str | None = None) -> dict[str, Any]:
        return {
            "NotebookExecution": {
                "Status": status,
                "LastStateChangeReason": reason,
            },
            "ResponseMetadata": {
                "HTTPStatusCode": 200,
            },
        }

    @mock.patch("airflow.providers.amazon.aws.hooks.emr.EmrHook.conn")
    def test_emr_notebook_execution_sensor_success_state(self, mock_conn):
        mock_conn.describe_notebook_execution.return_value = self._generate_response("FINISHED")
        sensor = EmrNotebookExecutionSensor(
            task_id="test_task",
            poke_interval=0,
            notebook_execution_id="test-execution-id",
        )
        sensor.poke(None)
        mock_conn.describe_notebook_execution.assert_called_once_with(NotebookExecutionId="test-execution-id")

    @mock.patch("airflow.providers.amazon.aws.hooks.emr.EmrHook.conn")
    def test_emr_notebook_execution_sensor_failed_state(self, mock_conn):
        error_reason = "Test error"
        mock_conn.describe_notebook_execution.return_value = self._generate_response("FAILED", error_reason)
        sensor = EmrNotebookExecutionSensor(
            task_id="test_task",
            poke_interval=0,
            notebook_execution_id="test-execution-id",
        )
        with pytest.raises(AirflowException, match=rf"EMR job failed: {error_reason}"):
            sensor.poke(None)
        mock_conn.describe_notebook_execution.assert_called_once_with(NotebookExecutionId="test-execution-id")

    @mock.patch("airflow.providers.amazon.aws.hooks.emr.EmrHook.conn")
    def test_emr_notebook_execution_sensor_success_state_multiple(self, mock_conn):
        return_values = [self._generate_response("PENDING") for i in range(2)]
        return_values.append(self._generate_response("FINISHED"))
        mock_conn.describe_notebook_execution.side_effect = return_values
        sensor = EmrNotebookExecutionSensor(
            task_id="test_task",
            poke_interval=0,
            notebook_execution_id="test-execution-id",
        )
        while not sensor.poke(None):
            pass
        assert mock_conn.describe_notebook_execution.call_count == 3



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Publish message to SQS queue."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.providers.amazon.aws.hooks.sqs import SqsHook
from airflow.providers.amazon.aws.operators.base_aws import AwsBaseOperator
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields

if TYPE_CHECKING:
    from airflow.utils.context import Context


class SqsPublishOperator(AwsBaseOperator[SqsHook]):
    """
    Publish a message to an Amazon SQS queue.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:SqsPublishOperator`

    :param sqs_queue: The SQS queue url (templated)
    :param message_content: The message content (templated)
    :param message_attributes: additional attributes for the message (default: None)
        For details of the attributes parameter see :py:meth:`botocore.client.SQS.send_message`
    :param delay_seconds: message delay (templated) (default: 1 second)
    :param message_group_id: This parameter applies only to FIFO (first-in-first-out) queues. (default: None)
        For details of the attributes parameter see :py:meth:`botocore.client.SQS.send_message`
    :param message_deduplication_id: This applies only to FIFO (first-in-first-out) queues.
        For details of the attributes parameter see :py:meth:`botocore.client.SQS.send_message`
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = SqsHook
    template_fields: Sequence[str] = aws_template_fields(
        "sqs_queue",
        "message_content",
        "delay_seconds",
        "message_attributes",
        "message_group_id",
        "message_deduplication_id",
    )
    template_fields_renderers = {"message_attributes": "json"}
    ui_color = "#6ad3fa"

    def __init__(
        self,
        *,
        sqs_queue: str,
        message_content: str,
        message_attributes: dict | None = None,
        delay_seconds: int = 0,
        message_group_id: str | None = None,
        message_deduplication_id: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.sqs_queue = sqs_queue
        self.message_content = message_content
        self.delay_seconds = delay_seconds
        self.message_attributes = message_attributes or {}
        self.message_group_id = message_group_id
        self.message_deduplication_id = message_deduplication_id

    def execute(self, context: Context) -> dict:
        """
        Publish the message to the Amazon SQS queue.

        :param context: the context object
        :return: dict with information about the message sent
            For details of the returned dict see :py:meth:`botocore.client.SQS.send_message`
        """
        result = self.hook.send_message(
            queue_url=self.sqs_queue,
            message_body=self.message_content,
            delay_seconds=self.delay_seconds,
            message_attributes=self.message_attributes,
            message_group_id=self.message_group_id,
            message_deduplication_id=self.message_deduplication_id,
        )

        self.log.info("send_message result: %s", result)

        return result



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from marshmallow import EXCLUDE, Schema, fields, post_load

from airflow.providers.amazon.aws.executors.batch.utils import BatchJob


class BatchSubmitJobResponseSchema(Schema):
    """API Response for SubmitJob."""

    # The unique identifier for the job.
    job_id = fields.String(data_key="jobId", required=True)

    class Meta:
        """Options object for a Schema. See Schema.Meta for more details and valid values."""

        unknown = EXCLUDE


class BatchJobDetailSchema(Schema):
    """API Response for Describe Jobs."""

    # The unique identifier for the job.
    job_id = fields.String(data_key="jobId", required=True)
    # The current status for the job:
    # 'SUBMITTED', 'PENDING', 'RUNNABLE', 'STARTING', 'RUNNING', 'SUCCEEDED', 'FAILED'
    status = fields.String(required=True)
    # A short, human-readable string to provide additional details about the current status of the job.
    status_reason = fields.String(data_key="statusReason")

    @post_load
    def make_job(self, data, **kwargs):
        """Overwrite marshmallow load() to return an instance of BatchJob instead of a dictionary."""
        return BatchJob(**data)

    class Meta:
        """Options object for a Schema. See Schema.Meta for more details and valid values."""

        unknown = EXCLUDE


class BatchDescribeJobsResponseSchema(Schema):
    """API Response for Describe Jobs."""

    # The list of jobs
    jobs = fields.List(fields.Nested(BatchJobDetailSchema), required=True)

    class Meta:
        """Options object for a Schema. See Schema.Meta for more details and valid values."""

        unknown = EXCLUDE



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import patch

import botocore.client
import pytest

from airflow.exceptions import AirflowException, TaskDeferred
from airflow.providers.amazon.aws.hooks.batch_client import BatchClientHook
from airflow.providers.amazon.aws.operators.batch import BatchCreateComputeEnvironmentOperator, BatchOperator

# Use dummy AWS credentials
from airflow.providers.amazon.aws.triggers.batch import (
    BatchCreateComputeEnvironmentTrigger,
    BatchJobTrigger,
)

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

AWS_REGION = "eu-west-1"
AWS_ACCESS_KEY_ID = "airflow_dummy_key"
AWS_SECRET_ACCESS_KEY = "airflow_dummy_secret"

JOB_NAME = "51455483-c62c-48ac-9b88-53a6a725baa3"
JOB_ID = "8ba9d676-4108-4474-9dca-8bbac1da9b19"

RESPONSE_WITHOUT_FAILURES = {
    "jobName": JOB_NAME,
    "jobId": JOB_ID,
}


class TestBatchOperator:
    MAX_RETRIES = 2
    STATUS_RETRIES = 3

    @mock.patch.dict("os.environ", AWS_DEFAULT_REGION=AWS_REGION)
    @mock.patch.dict("os.environ", AWS_ACCESS_KEY_ID=AWS_ACCESS_KEY_ID)
    @mock.patch.dict("os.environ", AWS_SECRET_ACCESS_KEY=AWS_SECRET_ACCESS_KEY)
    @mock.patch("airflow.providers.amazon.aws.hooks.batch_client.AwsBaseHook.get_client_type")
    def setup_method(self, _, get_client_type_mock):
        self.get_client_type_mock = get_client_type_mock
        self.batch = BatchOperator(
            task_id="task",
            job_name=JOB_NAME,
            job_queue="queue",
            job_definition="hello-world",
            max_retries=self.MAX_RETRIES,
            status_retries=self.STATUS_RETRIES,
            parameters=None,
            retry_strategy={"attempts": 1},
            container_overrides={},
            array_properties=None,
            aws_conn_id="airflow_test",
            region_name="eu-west-1",
            tags={},
            submit_job_timeout=3600,
        )
        self.client_mock = self.get_client_type_mock.return_value
        # We're mocking all actual AWS calls and don't need a connection. This
        # avoids an Airflow warning about connection cannot be found.
        self.batch.hook.get_connection = lambda _: None
        assert self.batch.hook.client == self.client_mock  # setup client property

        # don't pause in unit tests
        self.mock_delay = mock.Mock(return_value=None)
        self.batch.delay = self.mock_delay
        self.mock_exponential_delay = mock.Mock(return_value=0)
        self.batch.exponential_delay = self.mock_exponential_delay

        # Assign a job ID for most tests, so they don't depend on a job submission.
        assert self.batch.job_id is None
        self.batch.job_id = JOB_ID

        self.mock_context = mock.MagicMock()

    def test_init(self):
        assert self.batch.job_id == JOB_ID
        assert self.batch.job_name == JOB_NAME
        assert self.batch.job_queue == "queue"
        assert self.batch.job_definition == "hello-world"
        assert self.batch.waiters is None
        assert self.batch.hook.max_retries == self.MAX_RETRIES
        assert self.batch.hook.status_retries == self.STATUS_RETRIES
        assert self.batch.parameters == {}
        assert self.batch.retry_strategy == {"attempts": 1}
        assert self.batch.container_overrides == {}
        assert self.batch.array_properties is None
        assert self.batch.node_overrides is None
        assert self.batch.share_identifier is None
        assert self.batch.scheduling_priority_override is None
        assert self.batch.hook.region_name == "eu-west-1"
        assert self.batch.hook.aws_conn_id == "airflow_test"
        assert self.batch.hook.client == self.client_mock
        assert self.batch.tags == {}
        assert self.batch.wait_for_completion is True
        assert self.batch.submit_job_timeout == 3600

        self.get_client_type_mock.assert_called_once_with(region_name="eu-west-1")

    def test_init_defaults(self):
        """Test constructor default values"""
        batch_job = BatchOperator(
            task_id="task",
            job_name=JOB_NAME,
            job_queue="queue",
            job_definition="hello-world",
        )
        assert batch_job.job_id is None
        assert batch_job.job_name == JOB_NAME
        assert batch_job.job_queue == "queue"
        assert batch_job.job_definition == "hello-world"
        assert batch_job.waiters is None
        assert batch_job.hook.max_retries == 4200
        assert batch_job.hook.status_retries == 10
        assert batch_job.parameters == {}
        assert batch_job.retry_strategy is None
        assert batch_job.container_overrides is None
        assert batch_job.array_properties is None
        assert batch_job.ecs_properties_override is None
        assert batch_job.eks_properties_override is None
        assert batch_job.node_overrides is None
        assert batch_job.share_identifier is None
        assert batch_job.scheduling_priority_override is None
        assert batch_job.hook.region_name is None
        assert issubclass(type(batch_job.hook.client), botocore.client.BaseClient)
        assert batch_job.tags == {}
        assert batch_job.wait_for_completion is True
        assert batch_job.submit_job_timeout is None

    def test_template_fields_overrides(self):
        validate_template_fields(self.batch)

    @mock.patch.object(BatchClientHook, "get_job_description")
    @mock.patch.object(BatchClientHook, "wait_for_job")
    @mock.patch.object(BatchClientHook, "check_job_success")
    def test_execute_without_failures(self, check_mock, wait_mock, job_description_mock):
        # JOB_ID is in RESPONSE_WITHOUT_FAILURES
        self.client_mock.submit_job.return_value = RESPONSE_WITHOUT_FAILURES
        self.batch.job_id = None
        self.batch.waiters = None  # use default wait

        self.batch.execute(self.mock_context)

        self.client_mock.submit_job.assert_called_once_with(
            jobQueue="queue",
            jobName=JOB_NAME,
            containerOverrides={},
            jobDefinition="hello-world",
            parameters={},
            retryStrategy={"attempts": 1},
            tags={},
            timeout={"attemptDurationSeconds": 3600},
        )

        assert self.batch.job_id == JOB_ID
        wait_mock.assert_called_once_with(JOB_ID)
        check_mock.assert_called_once_with(JOB_ID)

        # First Call: Retrieve Batch Queue and Job Definition
        # Second Call: Retrieve CloudWatch information
        assert job_description_mock.call_count == 2

    def test_execute_with_failures(self):
        self.client_mock.submit_job.side_effect = Exception()

        with pytest.raises(AirflowException):
            self.batch.execute(self.mock_context)

        self.client_mock.submit_job.assert_called_once_with(
            jobQueue="queue",
            jobName=JOB_NAME,
            containerOverrides={},
            jobDefinition="hello-world",
            parameters={},
            retryStrategy={"attempts": 1},
            tags={},
            timeout={"attemptDurationSeconds": 3600},
        )

    @mock.patch.object(BatchClientHook, "get_job_description")
    @mock.patch.object(BatchClientHook, "wait_for_job")
    @mock.patch.object(BatchClientHook, "check_job_success")
    def test_execute_with_ecs_overrides(self, check_mock, wait_mock, job_description_mock):
        self.batch.container_overrides = None
        self.batch.ecs_properties_override = {
            "taskProperties": [
                {
                    "containers": [
                        {
                            "command": [
                                "string",
                            ],
                            "environment": [
                                {"name": "string", "value": "string"},
                            ],
                            "name": "string",
                            "resourceRequirements": [
                                {"value": "string", "type": "'GPU'|'VCPU'|'MEMORY'"},
                            ],
                        },
                    ]
                },
            ]
        }
        self.batch.execute(self.mock_context)

        self.client_mock.submit_job.assert_called_once_with(
            jobQueue="queue",
            jobName=JOB_NAME,
            jobDefinition="hello-world",
            ecsPropertiesOverride={
                "taskProperties": [
                    {
                        "containers": [
                            {
                                "command": [
                                    "string",
                                ],
                                "environment": [
                                    {"name": "string", "value": "string"},
                                ],
                                "name": "string",
                                "resourceRequirements": [
                                    {"value": "string", "type": "'GPU'|'VCPU'|'MEMORY'"},
                                ],
                            },
                        ]
                    },
                ]
            },
            parameters={},
            retryStrategy={"attempts": 1},
            tags={},
            timeout={"attemptDurationSeconds": 3600},
        )

    @mock.patch.object(BatchClientHook, "get_job_description")
    @mock.patch.object(BatchClientHook, "wait_for_job")
    @mock.patch.object(BatchClientHook, "check_job_success")
    def test_execute_with_eks_overrides(self, check_mock, wait_mock, job_description_mock):
        self.batch.container_overrides = None
        self.batch.eks_properties_override = {
            "podProperties": [
                {
                    "containers": [
                        {
                            "image": "string",
                            "command": [
                                "string",
                            ],
                            "args": [
                                "string",
                            ],
                            "env": [
                                {"name": "string", "value": "string"},
                            ],
                            "resources": [{"limits": {"string": "string"}, "requests": {"string": "string"}}],
                        },
                    ],
                    "initContainers": [
                        {
                            "image": "string",
                            "command": [
                                "string",
                            ],
                            "args": [
                                "string",
                            ],
                            "env": [
                                {"name": "string", "value": "string"},
                            ],
                            "resources": [{"limits": {"string": "string"}, "requests": {"string": "string"}}],
                        },
                    ],
                    "metadata": {
                        "labels": {"string": "string"},
                    },
                },
            ]
        }
        self.batch.execute(self.mock_context)

        self.client_mock.submit_job.assert_called_once_with(
            jobQueue="queue",
            jobName=JOB_NAME,
            jobDefinition="hello-world",
            eksPropertiesOverride={
                "podProperties": [
                    {
                        "containers": [
                            {
                                "image": "string",
                                "command": [
                                    "string",
                                ],
                                "args": [
                                    "string",
                                ],
                                "env": [
                                    {"name": "string", "value": "string"},
                                ],
                                "resources": [
                                    {"limits": {"string": "string"}, "requests": {"string": "string"}}
                                ],
                            },
                        ],
                        "initContainers": [
                            {
                                "image": "string",
                                "command": [
                                    "string",
                                ],
                                "args": [
                                    "string",
                                ],
                                "env": [
                                    {"name": "string", "value": "string"},
                                ],
                                "resources": [
                                    {"limits": {"string": "string"}, "requests": {"string": "string"}}
                                ],
                            },
                        ],
                        "metadata": {
                            "labels": {"string": "string"},
                        },
                    },
                ]
            },
            parameters={},
            retryStrategy={"attempts": 1},
            tags={},
            timeout={"attemptDurationSeconds": 3600},
        )

    @mock.patch.object(BatchClientHook, "check_job_success")
    def test_wait_job_complete_using_waiters(self, check_mock):
        mock_waiters = mock.Mock()
        self.batch.waiters = mock_waiters

        self.client_mock.submit_job.return_value = RESPONSE_WITHOUT_FAILURES
        self.client_mock.describe_jobs.return_value = {
            "jobs": [
                {
                    "jobId": JOB_ID,
                    "status": "SUCCEEDED",
                    "logStreamName": "logStreamName",
                    "container": {"logConfiguration": {}},
                }
            ]
        }
        self.batch.execute(self.mock_context)
        mock_waiters.wait_for_job.assert_called_once_with(JOB_ID)
        check_mock.assert_called_once_with(JOB_ID)

    @mock.patch.object(BatchClientHook, "check_job_success")
    def test_do_not_wait_job_complete(self, check_mock):
        self.batch.wait_for_completion = False

        self.client_mock.submit_job.return_value = RESPONSE_WITHOUT_FAILURES
        self.batch.execute(self.mock_context)

        check_mock.assert_not_called()

    def test_kill_job(self):
        self.client_mock.terminate_job.return_value = {}
        self.batch.on_kill()
        self.client_mock.terminate_job.assert_called_once_with(jobId=JOB_ID, reason="Task killed by the user")

    @pytest.mark.parametrize(
        "override", ["node_overrides", "ecs_properties_override", "eks_properties_override"]
    )
    @patch(
        "airflow.providers.amazon.aws.hooks.batch_client.BatchClientHook.client",
        new_callable=mock.PropertyMock,
    )
    def test_override_not_sent_if_not_set(self, client_mock, override):
        """
        check that when setting container override or node override, the other key is not sent
        in the API call (which would create a validation error from boto)
        """
        override_arg = {override: {"a": "a"}}
        batch = BatchOperator(
            task_id="task",
            job_name=JOB_NAME,
            job_queue="queue",
            job_definition="hello-world",
            **override_arg,
            # setting those to bypass code that is not relevant here
            do_xcom_push=False,
            wait_for_completion=False,
        )

        batch.execute(None)

        expected_args = {
            "jobQueue": "queue",
            "jobName": JOB_NAME,
            "jobDefinition": "hello-world",
            "parameters": {},
            "tags": {},
        }

        py2api = {
            "overrides": "containerOverrides",
            "node_overrides": "nodeOverrides",
            "ecs_properties_override": "ecsPropertiesOverride",
            "eks_properties_override": "eksPropertiesOverride",
        }

        expected_args[py2api[override]] = {"a": "a"}

        client_mock().submit_job.assert_called_once_with(**expected_args)

    def test_cant_set_old_and_new_override_param(self):
        with pytest.raises((TypeError, AirflowException), match="Invalid arguments were passed"):
            _ = BatchOperator(
                task_id="task",
                job_name=JOB_NAME,
                job_queue="queue",
                job_definition="hello-world",
                # can't set both of those, as one is a replacement for the other
                overrides={"a": "b"},
                container_overrides={"a": "b"},
            )

    @mock.patch.object(BatchClientHook, "get_job_description")
    @mock.patch("airflow.providers.amazon.aws.hooks.batch_client.AwsBaseHook.get_client_type")
    def test_defer_if_deferrable_param_set(self, mock_client, mock_get_job_description):
        mock_get_job_description.return_value = {"status": "SUBMITTED"}

        batch = BatchOperator(
            task_id="task",
            job_name=JOB_NAME,
            job_queue="queue",
            job_definition="hello-world",
            do_xcom_push=False,
            deferrable=True,
        )

        with pytest.raises(TaskDeferred) as exc:
            batch.execute(self.mock_context)
        assert isinstance(exc.value.trigger, BatchJobTrigger)

    @mock.patch("airflow.providers.amazon.aws.hooks.batch_client.AwsBaseHook.get_client_type")
    def test_defer_but_failed_due_to_job_id_not_found(self, mock_client):
        """Test that an AirflowException is raised if job_id is not set before deferral."""
        mock_client.return_value.submit_job.return_value = {
            "jobName": JOB_NAME,
            "jobId": None,
        }

        batch = BatchOperator(
            task_id="task",
            job_name=JOB_NAME,
            job_queue="queue",
            job_definition="hello-world",
            do_xcom_push=False,
            deferrable=True,
        )
        with pytest.raises(AirflowException) as exc:
            batch.execute(self.mock_context)
        assert "AWS Batch job - job_id was not found" in str(exc.value)

    @mock.patch.object(BatchClientHook, "get_job_description")
    @mock.patch("airflow.providers.amazon.aws.hooks.batch_client.AwsBaseHook.get_client_type")
    def test_defer_but_success_before_deferred(self, mock_client, mock_get_job_description):
        """Test that an AirflowException is raised if job_id is not set before deferral."""
        mock_client.return_value.submit_job.return_value = RESPONSE_WITHOUT_FAILURES
        mock_get_job_description.return_value = {"status": "SUCCEEDED"}

        batch = BatchOperator(
            task_id="task",
            job_name=JOB_NAME,
            job_queue="queue",
            job_definition="hello-world",
            do_xcom_push=False,
            deferrable=True,
        )
        assert batch.execute(self.mock_context) == JOB_ID

    @mock.patch.object(BatchClientHook, "get_job_description")
    @mock.patch("airflow.providers.amazon.aws.hooks.batch_client.AwsBaseHook.get_client_type")
    def test_defer_but_fail_before_deferred(self, mock_client, mock_get_job_description):
        """Test that an AirflowException is raised if job_id is not set before deferral."""
        mock_client.return_value.submit_job.return_value = RESPONSE_WITHOUT_FAILURES
        mock_get_job_description.return_value = {"status": "FAILED"}

        batch = BatchOperator(
            task_id="task",
            job_name=JOB_NAME,
            job_queue="queue",
            job_definition="hello-world",
            do_xcom_push=False,
            deferrable=True,
        )
        with pytest.raises(AirflowException) as exc:
            batch.execute(self.mock_context)
        assert f"Error while running job: {JOB_ID} is in FAILED state" in str(exc.value)

    @mock.patch.object(BatchClientHook, "get_job_description")
    @mock.patch.object(BatchClientHook, "wait_for_job")
    @mock.patch.object(BatchClientHook, "check_job_success")
    @mock.patch("airflow.providers.amazon.aws.links.batch.BatchJobQueueLink.persist")
    @mock.patch("airflow.providers.amazon.aws.links.batch.BatchJobDefinitionLink.persist")
    def test_monitor_job_with_logs(
        self, job_definition_persist_mock, job_queue_persist_mock, check_mock, wait_mock, job_description_mock
    ):
        batch = BatchOperator(
            task_id="task",
            job_name=JOB_NAME,
            job_queue="queue",
            job_definition="hello-world",
            awslogs_enabled=True,
        )

        batch.job_id = JOB_ID

        batch.monitor_job(context=None)

        job_description_mock.assert_called_with(job_id=JOB_ID)
        job_definition_persist_mock.assert_called_once()
        job_queue_persist_mock.assert_called_once()
        wait_mock.assert_called_once()
        assert len(wait_mock.call_args) == 2


class TestBatchCreateComputeEnvironmentOperator:
    warn_message = "The `status_retries` parameter is unused and should be removed"

    def test_template_fields(self):
        environment_name = "environment_name"
        compute_resources = {}
        service_role = "test_role"
        environment_type = "environment_type"
        environment_state = "state"

        operator = BatchCreateComputeEnvironmentOperator(
            task_id="task",
            compute_environment_name=environment_name,
            environment_type=environment_type,
            state=environment_state,
            compute_resources=compute_resources,
            service_role=service_role,
        )

        validate_template_fields(operator)

    @mock.patch.object(BatchClientHook, "client")
    def test_execute(self, mock_conn):
        environment_name = "environment_name"
        environment_type = "environment_type"
        environment_state = "environment_state"
        compute_resources = {}
        tags = {}
        operator = BatchCreateComputeEnvironmentOperator(
            task_id="task",
            compute_environment_name=environment_name,
            environment_type=environment_type,
            state=environment_state,
            compute_resources=compute_resources,
            tags=tags,
        )
        operator.execute(None)
        mock_conn.create_compute_environment.assert_called_once_with(
            computeEnvironmentName=environment_name,
            type=environment_type,
            state=environment_state,
            computeResources=compute_resources,
            tags=tags,
        )

    @mock.patch.object(BatchClientHook, "client")
    def test_defer(self, client_mock):
        client_mock.create_compute_environment.return_value = {"computeEnvironmentArn": "my_arn"}

        operator = BatchCreateComputeEnvironmentOperator(
            task_id="task",
            compute_environment_name="my_env_name",
            environment_type="my_env_type",
            state="my_state",
            compute_resources={},
            max_retries=123456,
            poll_interval=456789,
            deferrable=True,
        )

        with pytest.raises(TaskDeferred) as deferred:
            operator.execute(None)

        assert isinstance(deferred.value.trigger, BatchCreateComputeEnvironmentTrigger)
        assert deferred.value.trigger.waiter_delay == 456789
        assert deferred.value.trigger.attempts == 123456



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime

import pytest

from airflow.providers.amazon.aws.utils import (
    _StringCompareEnum,
    datetime_to_epoch,
    datetime_to_epoch_ms,
    datetime_to_epoch_us,
    get_airflow_version,
    get_botocore_version,
)

DT = datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc)
EPOCH = 946_684_800


class EnumTest(_StringCompareEnum):
    FOO = "FOO"


def test_datetime_to_epoch():
    assert datetime_to_epoch(DT) == EPOCH


def test_datetime_to_epoch_ms():
    assert datetime_to_epoch_ms(DT) == EPOCH * 1000


def test_datetime_to_epoch_us():
    assert datetime_to_epoch_us(DT) == EPOCH * 1_000_000


def test_get_airflow_version():
    assert len(get_airflow_version()) == 3


def test_str_enum():
    assert EnumTest.FOO == "FOO"
    assert EnumTest.FOO.value == "FOO"


def test_botocore_version():
    pytest.importorskip("botocore", reason="`botocore` not installed")

    botocore_version = get_botocore_version()
    assert len(botocore_version) == 3
    assert isinstance(botocore_version[0], int), "botocore major version expected to be an integer"
    assert isinstance(botocore_version[1], int), "botocore minor version expected to be an integer"
    assert isinstance(botocore_version[2], int), "botocore patch version expected to be an integer"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.dms import DmsHook
from airflow.providers.amazon.aws.sensors.dms import DmsTaskCompletedSensor


@pytest.fixture
def mocked_get_task_status():
    with mock.patch.object(DmsHook, "get_task_status") as m:
        yield m


class TestDmsTaskCompletedSensor:
    def setup_method(self):
        self.default_op_kwargs = {
            "task_id": "test_dms_sensor",
            "aws_conn_id": None,
            "replication_task_arn": "task_arn",
        }

    def test_init(self):
        self.default_op_kwargs.pop("aws_conn_id", None)

        sensor = DmsTaskCompletedSensor(
            **self.default_op_kwargs,
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="ca-west-1",
            verify=True,
            botocore_config={"read_timeout": 42},
        )
        assert sensor.hook.client_type == "dms"
        assert sensor.hook.resource_type is None
        assert sensor.hook.aws_conn_id == "fake-conn-id"
        assert sensor.hook._region_name == "ca-west-1"
        assert sensor.hook._verify is True
        assert sensor.hook._config is not None
        assert sensor.hook._config.read_timeout == 42

        sensor = DmsTaskCompletedSensor(task_id="create_task", replication_task_arn="task_arn")
        assert sensor.hook.aws_conn_id == "aws_default"
        assert sensor.hook._region_name is None
        assert sensor.hook._verify is None
        assert sensor.hook._config is None

    @pytest.mark.parametrize("status", ["stopped"])
    def test_poke_completed(self, mocked_get_task_status, status):
        mocked_get_task_status.return_value = status
        assert DmsTaskCompletedSensor(**self.default_op_kwargs).poke({})

    @pytest.mark.parametrize("status", ["running", "starting"])
    def test_poke_not_completed(self, mocked_get_task_status, status):
        mocked_get_task_status.return_value = status
        assert not DmsTaskCompletedSensor(**self.default_op_kwargs).poke({})

    @pytest.mark.parametrize(
        "status",
        [
            "creating",
            "deleting",
            "failed",
            "failed-move",
            "modifying",
            "moving",
            "ready",
            "testing",
        ],
    )
    def test_poke_terminated_status(self, mocked_get_task_status, status):
        mocked_get_task_status.return_value = status
        error_message = f"Unexpected status: {status}"
        with pytest.raises(AirflowException, match=error_message):
            DmsTaskCompletedSensor(**self.default_op_kwargs).poke({})

    def test_poke_none_status(self, mocked_get_task_status):
        mocked_get_task_status.return_value = None
        with pytest.raises(AirflowException, match="task with ARN .* not found"):
            DmsTaskCompletedSensor(**self.default_op_kwargs).poke({})



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow.decorators import task
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.hooks.sqs import SqsHook
from airflow.providers.amazon.aws.operators.sqs import SqsPublishOperator
from airflow.providers.amazon.aws.sensors.sqs import SqsSensor
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import SystemTestContextBuilder

sys_test_context_task = SystemTestContextBuilder().build()

DAG_ID = "example_sqs"


@task
def create_queue(queue_name) -> str:
    return SqsHook().create_queue(queue_name=queue_name)["QueueUrl"]


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_queue(queue_url):
    SqsHook().conn.delete_queue(QueueUrl=queue_url)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]

    sns_queue_name = f"{env_id}-example-queue"

    sqs_queue = create_queue(sns_queue_name)

    # [START howto_operator_sqs]
    publish_to_queue_1 = SqsPublishOperator(
        task_id="publish_to_queue_1",
        sqs_queue=sqs_queue,
        message_content="{{ task_instance }}",
    )
    publish_to_queue_2 = SqsPublishOperator(
        task_id="publish_to_queue_2",
        sqs_queue=sqs_queue,
        message_content="{{ task_instance }}",
    )
    # [END howto_operator_sqs]

    # [START howto_sensor_sqs]
    read_from_queue = SqsSensor(
        task_id="read_from_queue",
        sqs_queue=sqs_queue,
    )
    # Retrieve multiple batches of messages from SQS.
    # The SQS API only returns a maximum of 10 messages per poll.
    read_from_queue_in_batch = SqsSensor(
        task_id="read_from_queue_in_batch",
        sqs_queue=sqs_queue,
        # Get maximum 10 messages each poll
        max_messages=10,
        # Combine 3 polls before returning results
        num_batches=3,
    )
    # [END howto_sensor_sqs]

    chain(
        # TEST SETUP
        test_context,
        sqs_queue,
        # TEST BODY
        publish_to_queue_1,
        read_from_queue,
        publish_to_queue_2,
        read_from_queue_in_batch,
        # TEST TEARDOWN
        delete_queue(sqs_queue),
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.kinesis_analytics import KinesisAnalyticsV2Hook
from airflow.providers.amazon.aws.sensors.kinesis_analytics import (
    KinesisAnalyticsV2StartApplicationCompletedSensor,
    KinesisAnalyticsV2StopApplicationCompletedSensor,
)


class TestKinesisAnalyticsV2StartApplicationCompletedSensor:
    SENSOR = KinesisAnalyticsV2StartApplicationCompletedSensor
    APPLICATION_ARN = "arn:aws:kinesisanalytics:us-east-1:123456789012:application/demo"

    def setup_method(self):
        self.default_op_kwargs = dict(
            task_id="start_application_sensor",
            application_name="demo",
            poke_interval=5,
            max_retries=1,
        )
        self.sensor = self.SENSOR(**self.default_op_kwargs, aws_conn_id=None)

    def test_base_aws_op_attributes(self):
        op = self.SENSOR(**self.default_op_kwargs)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

        op = self.SENSOR(
            **self.default_op_kwargs,
            aws_conn_id="aws-test-custom-conn",
            region_name="eu-west-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.hook.aws_conn_id == "aws-test-custom-conn"
        assert op.hook._region_name == "eu-west-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

    @pytest.mark.parametrize("state", SENSOR.SUCCESS_STATES)
    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_poke_success_state(self, mock_conn, state):
        mock_conn.describe_application.return_value = {
            "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN, "ApplicationStatus": state}
        }

        assert self.sensor.poke({}) is True

    @pytest.mark.parametrize("state", SENSOR.INTERMEDIATE_STATES)
    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_intermediate_state(self, mock_conn, state):
        mock_conn.describe_application.return_value = {
            "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN, "ApplicationStatus": state}
        }
        assert self.sensor.poke({}) is False

    @pytest.mark.parametrize("state", SENSOR.FAILURE_STATES)
    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_poke_failure_states(self, mock_conn, state):
        mock_conn.describe_application.return_value = {
            "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN, "ApplicationStatus": state}
        }

        sensor = self.SENSOR(**self.default_op_kwargs, aws_conn_id=None)

        with pytest.raises(
            AirflowException, match="AWS Managed Service for Apache Flink application start failed"
        ):
            sensor.poke({})


class TestKinesisAnalyticsV2StopApplicationCompletedSensor:
    SENSOR = KinesisAnalyticsV2StopApplicationCompletedSensor
    APPLICATION_ARN = "arn:aws:kinesisanalytics:us-east-1:123456789012:application/demo"

    def setup_method(self):
        self.default_op_kwargs = dict(
            task_id="stop_application_sensor",
            application_name="demo",
            poke_interval=5,
            max_retries=1,
        )
        self.sensor = self.SENSOR(**self.default_op_kwargs, aws_conn_id=None)

    def test_base_aws_op_attributes(self):
        op = self.SENSOR(**self.default_op_kwargs)
        assert op.hook.aws_conn_id == "aws_default"
        assert op.hook._region_name is None
        assert op.hook._verify is None
        assert op.hook._config is None

        op = self.SENSOR(
            **self.default_op_kwargs,
            aws_conn_id="aws-test-custom-conn",
            region_name="eu-west-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert op.hook.aws_conn_id == "aws-test-custom-conn"
        assert op.hook._region_name == "eu-west-1"
        assert op.hook._verify is False
        assert op.hook._config is not None
        assert op.hook._config.read_timeout == 42

    @pytest.mark.parametrize("state", SENSOR.SUCCESS_STATES)
    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_poke_success_state(self, mock_conn, state):
        mock_conn.describe_application.return_value = {
            "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN, "ApplicationStatus": state}
        }

        assert self.sensor.poke({}) is True

    @pytest.mark.parametrize("state", SENSOR.INTERMEDIATE_STATES)
    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_intermediate_state(self, mock_conn, state):
        mock_conn.describe_application.return_value = {
            "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN, "ApplicationStatus": state}
        }
        assert self.sensor.poke({}) is False

    @pytest.mark.parametrize("state", SENSOR.FAILURE_STATES)
    @mock.patch.object(KinesisAnalyticsV2Hook, "conn")
    def test_poke_failure_states(self, mock_conn, state):
        mock_conn.describe_application.return_value = {
            "ApplicationDetail": {"ApplicationARN": self.APPLICATION_ARN, "ApplicationStatus": state}
        }

        sensor = self.SENSOR(**self.default_op_kwargs, aws_conn_id=None)

        with pytest.raises(
            AirflowException, match="AWS Managed Service for Apache Flink application stop failed"
        ):
            sensor.poke({})



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from copy import deepcopy
from unittest import mock

import pytest
from boto3.session import Session

from airflow.exceptions import AirflowException
from airflow.models.connection import Connection
from airflow.providers.amazon.aws.transfers.redshift_to_s3 import RedshiftToS3Operator
from airflow.providers.amazon.aws.utils.redshift import build_credentials_block
from airflow.providers.common.compat.openlineage.facet import (
    ColumnLineageDatasetFacet,
    DocumentationDatasetFacet,
    Fields,
    InputField,
    SchemaDatasetFacet,
    SchemaDatasetFacetFields,
)

from tests_common.test_utils.asserts import assert_equal_ignore_multiple_spaces


class TestRedshiftToS3Transfer:
    @pytest.mark.parametrize("table_as_file_name, expected_s3_key", [[True, "key/table_"], [False, "key"]])
    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.get_connection")
    @mock.patch("airflow.models.connection.Connection")
    @mock.patch("boto3.session.Session")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.RedshiftSQLHook.run")
    def test_table_unloading(
        self,
        mock_run,
        mock_session,
        mock_connection,
        mock_hook,
        table_as_file_name,
        expected_s3_key,
    ):
        access_key = "aws_access_key_id"
        secret_key = "aws_secret_access_key"
        mock_session.return_value = Session(access_key, secret_key)
        mock_session.return_value.access_key = access_key
        mock_session.return_value.secret_key = secret_key
        mock_session.return_value.token = None
        mock_connection.return_value = Connection()
        mock_hook.return_value = Connection()
        schema = "schema"
        table = "table"
        s3_bucket = "bucket"
        s3_key = "key"
        unload_options = [
            "HEADER",
        ]

        op = RedshiftToS3Operator(
            schema=schema,
            table=table,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            unload_options=unload_options,
            include_header=True,
            redshift_conn_id="redshift_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            table_as_file_name=table_as_file_name,
            dag=None,
        )

        op.execute(None)

        unload_options = "\n\t\t\t".join(unload_options)
        select_query = f"SELECT * FROM {schema}.{table}"
        credentials_block = build_credentials_block(mock_session.return_value)

        unload_query = op._build_unload_query(
            credentials_block, select_query, expected_s3_key, unload_options
        )

        assert mock_run.call_count == 1
        assert access_key in unload_query
        assert secret_key in unload_query
        assert_equal_ignore_multiple_spaces(mock_run.call_args.args[0], unload_query)

    @pytest.mark.parametrize("table_as_file_name, expected_s3_key", [[True, "key/table_"], [False, "key"]])
    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.get_connection")
    @mock.patch("airflow.models.connection.Connection")
    @mock.patch("boto3.session.Session")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.RedshiftSQLHook.run")
    def test_execute_sts_token(
        self,
        mock_run,
        mock_session,
        mock_connection,
        mock_hook,
        table_as_file_name,
        expected_s3_key,
    ):
        access_key = "ASIA_aws_access_key_id"
        secret_key = "aws_secret_access_key"
        token = "token"
        mock_session.return_value = Session(access_key, secret_key, token)
        mock_session.return_value.access_key = access_key
        mock_session.return_value.secret_key = secret_key
        mock_session.return_value.token = token
        mock_connection.return_value = Connection()
        mock_hook.return_value = Connection()
        schema = "schema"
        table = "table"
        s3_bucket = "bucket"
        s3_key = "key"
        unload_options = [
            "HEADER",
        ]

        op = RedshiftToS3Operator(
            schema=schema,
            table=table,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            unload_options=unload_options,
            include_header=True,
            redshift_conn_id="redshift_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            table_as_file_name=table_as_file_name,
            dag=None,
        )

        op.execute(None)

        unload_options = "\n\t\t\t".join(unload_options)
        select_query = f"SELECT * FROM {schema}.{table}"
        credentials_block = build_credentials_block(mock_session.return_value)

        unload_query = op._build_unload_query(
            credentials_block, select_query, expected_s3_key, unload_options
        )

        assert mock_run.call_count == 1
        assert access_key in unload_query
        assert secret_key in unload_query
        assert token in unload_query
        assert_equal_ignore_multiple_spaces(mock_run.call_args.args[0], unload_query)

    @pytest.mark.parametrize(
        "table, table_as_file_name, expected_s3_key",
        [
            ["table", True, "key/table_"],
            ["table", False, "key"],
            [None, False, "key"],
            [None, True, "key"],
        ],
    )
    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.get_connection")
    @mock.patch("airflow.models.connection.Connection")
    @mock.patch("boto3.session.Session")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.RedshiftSQLHook.run")
    def test_custom_select_query_unloading(
        self,
        mock_run,
        mock_session,
        mock_connection,
        mock_hook,
        table,
        table_as_file_name,
        expected_s3_key,
    ):
        access_key = "aws_access_key_id"
        secret_key = "aws_secret_access_key"
        mock_session.return_value = Session(access_key, secret_key)
        mock_session.return_value.access_key = access_key
        mock_session.return_value.secret_key = secret_key
        mock_session.return_value.token = None
        mock_connection.return_value = Connection()
        mock_hook.return_value = Connection()
        s3_bucket = "bucket"
        s3_key = "key"
        unload_options = [
            "HEADER",
        ]
        select_query = "select column from table"

        op = RedshiftToS3Operator(
            select_query=select_query,
            table=table,
            table_as_file_name=table_as_file_name,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            unload_options=unload_options,
            include_header=True,
            redshift_conn_id="redshift_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            dag=None,
        )

        op.execute(None)

        unload_options = "\n\t\t\t".join(unload_options)
        credentials_block = build_credentials_block(mock_session.return_value)

        unload_query = op._build_unload_query(
            credentials_block, select_query, expected_s3_key, unload_options
        )

        assert mock_run.call_count == 1
        assert access_key in unload_query
        assert secret_key in unload_query
        assert_equal_ignore_multiple_spaces(mock_run.call_args.args[0], unload_query)

    @pytest.mark.parametrize(
        "table_as_file_name, expected_s3_key, select_query, expected_query",
        [
            [
                True,
                "key/table_",
                "SELECT 'Single Quotes Break this Operator'",
                "SELECT 'Single Quotes Break this Operator'",
            ],
            [
                False,
                "key",
                "SELECT 'Single Quotes Break this Operator'",
                "SELECT 'Single Quotes Break this Operator'",
            ],
            [
                True,
                "key/table_",
                "SELECT ''Single Quotes Break this Operator''",
                "SELECT 'Single Quotes Break this Operator'",
            ],
            [
                False,
                "key",
                "SELECT ''Single Quotes Break this Operator''",
                "SELECT 'Single Quotes Break this Operator'",
            ],
            [False, "key", "SELECT ''", "SELECT ''"],
            [
                False,
                "key",
                "SELECT ''Single Quotes '' || ''Break this Operator''",
                "SELECT 'Single Quotes ' || 'Break this Operator'",
            ],
        ],
    )
    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.get_connection")
    @mock.patch("airflow.models.connection.Connection")
    @mock.patch("boto3.session.Session")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.RedshiftSQLHook.run")
    def test_custom_select_query_unloading_with_single_quotes(
        self,
        mock_run,
        mock_session,
        mock_connection,
        mock_hook,
        table_as_file_name,
        expected_s3_key,
        select_query,
        expected_query,
    ):
        access_key = "aws_access_key_id"
        secret_key = "aws_secret_access_key"
        mock_session.return_value = Session(access_key, secret_key)
        mock_session.return_value.access_key = access_key
        mock_session.return_value.secret_key = secret_key
        mock_session.return_value.token = None
        mock_connection.return_value = Connection()
        mock_hook.return_value = Connection()
        s3_bucket = "bucket"
        s3_key = "key"
        unload_options = ["HEADER"]

        op = RedshiftToS3Operator(
            select_query=select_query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            unload_options=unload_options,
            include_header=True,
            redshift_conn_id="redshift_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            dag=None,
        )

        op.execute(None)

        unload_options = "\n\t\t\t".join(unload_options)
        credentials_block = build_credentials_block(mock_session.return_value)

        unload_query = op._build_unload_query(credentials_block, select_query, s3_key, unload_options)

        assert mock_run.call_count == 1
        assert access_key in unload_query
        assert secret_key in unload_query
        assert_equal_ignore_multiple_spaces(mock_run.call_args.args[0], unload_query)
        assert f"UNLOAD ($${expected_query}$$)" in unload_query

    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.get_connection")
    @mock.patch("airflow.models.connection.Connection")
    @mock.patch("boto3.session.Session")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.RedshiftSQLHook.run")
    def test_custom_select_query_has_precedence_over_table_and_schema(
        self,
        mock_run,
        mock_session,
        mock_connection,
        mock_hook,
    ):
        access_key = "aws_access_key_id"
        secret_key = "aws_secret_access_key"
        mock_session.return_value = Session(access_key, secret_key)
        mock_session.return_value.access_key = access_key
        mock_session.return_value.secret_key = secret_key
        mock_session.return_value.token = None
        mock_connection.return_value = Connection()
        mock_hook.return_value = Connection()
        s3_bucket = "bucket"
        s3_key = "key"
        unload_options = [
            "HEADER",
        ]
        select_query = "select column from table"

        op = RedshiftToS3Operator(
            select_query=select_query,
            table="table",
            schema="schema",
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            unload_options=unload_options,
            include_header=True,
            redshift_conn_id="redshift_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            dag=None,
        )

        op.execute(None)

        unload_options = "\n\t\t\t".join(unload_options)
        credentials_block = build_credentials_block(mock_session.return_value)

        unload_query = op._build_unload_query(credentials_block, select_query, "key/table_", unload_options)

        assert mock_run.call_count == 1
        assert access_key in unload_query
        assert secret_key in unload_query
        assert_equal_ignore_multiple_spaces(mock_run.call_args.args[0], unload_query)

    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.get_connection")
    @mock.patch("airflow.models.connection.Connection")
    @mock.patch("boto3.session.Session")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.RedshiftSQLHook.run")
    def test_default_select_query_used_when_table_and_schema_missing(
        self,
        mock_run,
        mock_session,
        mock_connection,
        mock_hook,
    ):
        access_key = "aws_access_key_id"
        secret_key = "aws_secret_access_key"
        mock_session.return_value = Session(access_key, secret_key)
        mock_session.return_value.access_key = access_key
        mock_session.return_value.secret_key = secret_key
        mock_session.return_value.token = None
        mock_connection.return_value = Connection()
        mock_hook.return_value = Connection()
        s3_bucket = "bucket"
        s3_key = "key"
        unload_options = [
            "HEADER",
        ]
        default_query = "SELECT * FROM schema.table"

        op = RedshiftToS3Operator(
            table="table",
            schema="schema",
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            unload_options=unload_options,
            include_header=True,
            redshift_conn_id="redshift_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            dag=None,
        )

        op.execute(None)

        unload_options = "\n\t\t\t".join(unload_options)
        credentials_block = build_credentials_block(mock_session.return_value)

        unload_query = op._build_unload_query(credentials_block, default_query, "key/table_", unload_options)

        assert mock_run.call_count == 1
        assert access_key in unload_query
        assert secret_key in unload_query
        assert_equal_ignore_multiple_spaces(mock_run.call_args.args[0], unload_query)

    def test_lack_of_select_query_and_schema_and_table_raises_error(self):
        op = RedshiftToS3Operator(
            s3_bucket="bucket",
            s3_key="key",
            include_header=True,
            redshift_conn_id="redshift_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            dag=None,
        )

        with pytest.raises(ValueError):
            op.execute(None)

    @pytest.mark.parametrize("table_as_file_name, expected_s3_key", [[True, "key/table_"], [False, "key"]])
    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.get_connection")
    @mock.patch("airflow.models.connection.Connection")
    @mock.patch("boto3.session.Session")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.RedshiftSQLHook.run")
    def test_table_unloading_role_arn(
        self,
        mock_run,
        mock_session,
        mock_connection,
        mock_hook,
        table_as_file_name,
        expected_s3_key,
    ):
        access_key = "aws_access_key_id"
        secret_key = "aws_secret_access_key"
        extra = {"role_arn": "arn:aws:iam::112233445566:role/myRole"}
        mock_session.return_value = Session(access_key, secret_key)
        mock_session.return_value.access_key = access_key
        mock_session.return_value.secret_key = secret_key
        mock_session.return_value.token = None
        mock_connection.return_value = Connection(extra=extra)
        mock_hook.return_value = Connection(extra=extra)
        schema = "schema"
        table = "table"
        s3_bucket = "bucket"
        s3_key = "key"
        unload_options = [
            "HEADER",
        ]

        op = RedshiftToS3Operator(
            schema=schema,
            table=table,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            unload_options=unload_options,
            include_header=True,
            redshift_conn_id="redshift_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            table_as_file_name=table_as_file_name,
            dag=None,
        )

        op.execute(None)

        unload_options = "\n\t\t\t".join(unload_options)
        select_query = f"SELECT * FROM {schema}.{table}"
        credentials_block = f"aws_iam_role={extra['role_arn']}"

        unload_query = op._build_unload_query(
            credentials_block, select_query, expected_s3_key, unload_options
        )

        assert mock_run.call_count == 1
        assert extra["role_arn"] in unload_query
        assert_equal_ignore_multiple_spaces(mock_run.call_args.args[0], unload_query)

    @pytest.mark.parametrize("param", ["sql", "parameters"])
    def test_invalid_param_in_redshift_data_api_kwargs(self, param):
        """
        Test passing invalid param in RS Data API kwargs raises an error
        """
        redshift_operator = RedshiftToS3Operator(
            s3_bucket="s3_bucket",
            s3_key="s3_key",
            select_query="select_query",
            task_id="task_id",
            dag=None,
            redshift_data_api_kwargs={param: "param"},
        )
        with pytest.raises(AirflowException):
            redshift_operator.execute(None)

    @pytest.mark.parametrize("table_as_file_name, expected_s3_key", [[True, "key/table_"], [False, "key"]])
    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.get_connection")
    @mock.patch("airflow.models.connection.Connection")
    @mock.patch("boto3.session.Session")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.RedshiftSQLHook.run")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_data.RedshiftDataHook.conn")
    def test_table_unloading_using_redshift_data_api(
        self,
        mock_rs,
        mock_run,
        mock_session,
        mock_connection,
        mock_hook,
        table_as_file_name,
        expected_s3_key,
    ):
        access_key = "aws_access_key_id"
        secret_key = "aws_secret_access_key"
        mock_session.return_value = Session(access_key, secret_key)
        mock_session.return_value.access_key = access_key
        mock_session.return_value.secret_key = secret_key
        mock_session.return_value.token = None
        mock_connection.return_value = Connection()
        mock_hook.return_value = Connection()
        mock_rs.execute_statement.return_value = {"Id": "STATEMENT_ID"}
        mock_rs.describe_statement.return_value = {"Status": "FINISHED"}

        schema = "schema"
        table = "table"
        s3_bucket = "bucket"
        s3_key = "key"
        unload_options = [
            "HEADER",
        ]
        # RS Data API params
        database = "database"
        cluster_identifier = "cluster_identifier"
        db_user = "db_user"
        secret_arn = "secret_arn"
        statement_name = "statement_name"

        op = RedshiftToS3Operator(
            schema=schema,
            table=table,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            unload_options=unload_options,
            include_header=True,
            redshift_conn_id="redshift_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            table_as_file_name=table_as_file_name,
            dag=None,
            redshift_data_api_kwargs=dict(
                database=database,
                cluster_identifier=cluster_identifier,
                db_user=db_user,
                secret_arn=secret_arn,
                statement_name=statement_name,
            ),
        )

        op.execute(None)

        unload_options = "\n\t\t\t".join(unload_options)
        select_query = f"SELECT * FROM {schema}.{table}"
        credentials_block = build_credentials_block(mock_session.return_value)

        unload_query = op._build_unload_query(
            credentials_block, select_query, expected_s3_key, unload_options
        )

        mock_run.assert_not_called()
        assert access_key in unload_query
        assert secret_key in unload_query

        mock_rs.execute_statement.assert_called_once()
        # test with all args besides sql
        _call = deepcopy(mock_rs.execute_statement.call_args.kwargs)
        _call.pop("Sql")
        assert _call == dict(
            Database=database,
            ClusterIdentifier=cluster_identifier,
            DbUser=db_user,
            SecretArn=secret_arn,
            StatementName=statement_name,
            WithEvent=False,
        )
        mock_rs.describe_statement.assert_called_once_with(
            Id="STATEMENT_ID",
        )
        # test sql arg
        assert_equal_ignore_multiple_spaces(mock_rs.execute_statement.call_args.kwargs["Sql"], unload_query)

    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.get_connection")
    @mock.patch("airflow.models.connection.Connection.get_connection_from_secrets")
    @mock.patch("boto3.session.Session")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.RedshiftSQLHook.run")
    @mock.patch("airflow.providers.amazon.aws.utils.openlineage.get_facets_from_redshift_table")
    def test_get_openlineage_facets_on_complete_default(
        self, mock_get_facets, mock_run, mock_session, mock_connection, mock_hook
    ):
        access_key = "aws_access_key_id"
        secret_key = "aws_secret_access_key"
        mock_session.return_value = Session(access_key, secret_key)
        mock_session.return_value.access_key = access_key
        mock_session.return_value.secret_key = secret_key
        mock_session.return_value.token = None

        mock_connection.return_value = mock.MagicMock(
            schema="database", port=5439, host="cluster.id.region.redshift.amazonaws.com", extra_dejson={}
        )
        mock_facets = {
            "schema": SchemaDatasetFacet(fields=[SchemaDatasetFacetFields(name="col", type="STRING")]),
            "documentation": DocumentationDatasetFacet(description="mock_description"),
        }
        mock_get_facets.return_value = mock_facets

        schema = "schema"
        table = "table"
        s3_bucket = "bucket"
        s3_key = "key"

        op = RedshiftToS3Operator(
            schema=schema,
            table=table,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            redshift_conn_id="redshift_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
        )
        op.execute(None)

        lineage = op.get_openlineage_facets_on_complete(None)
        # Hook called only one time - on operator execution - we mocked querying to fetch schema
        assert mock_run.call_count == 1

        assert len(lineage.inputs) == 1
        assert len(lineage.outputs) == 1
        assert lineage.outputs[0].name == f"{s3_key}/{table}_"
        assert lineage.outputs[0].namespace == f"s3://{s3_bucket}"
        assert lineage.inputs[0].name == f"database.{schema}.{table}"
        assert lineage.inputs[0].namespace == "redshift://cluster.region:5439"

        assert lineage.inputs[0].facets == mock_facets
        assert lineage.outputs[0].facets == {
            "columnLineage": ColumnLineageDatasetFacet(
                fields={
                    "col": Fields(
                        inputFields=[
                            InputField(
                                namespace="redshift://cluster.region:5439",
                                name=f"database.{schema}.{table}",
                                field="col",
                            )
                        ],
                        transformationType="IDENTITY",
                        transformationDescription="identical",
                    )
                }
            ),
            "schema": SchemaDatasetFacet(fields=[SchemaDatasetFacetFields(name="col", type="STRING")]),
        }

    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.get_connection")
    @mock.patch("airflow.models.connection.Connection.get_connection_from_secrets")
    @mock.patch("boto3.session.Session")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.RedshiftSQLHook.run")
    @mock.patch("airflow.providers.amazon.aws.utils.openlineage.get_facets_from_redshift_table")
    def test_get_openlineage_facets_on_complete_with_select_query(
        self, mock_get_facets, mock_run, mock_session, mock_connection, mock_hook
    ):
        access_key = "aws_access_key_id"
        secret_key = "aws_secret_access_key"
        mock_session.return_value = Session(access_key, secret_key)
        mock_session.return_value.access_key = access_key
        mock_session.return_value.secret_key = secret_key
        mock_session.return_value.token = None

        mock_connection.return_value = mock.MagicMock(
            schema="database", port=5439, host="cluster.id.region.redshift.amazonaws.com", extra_dejson={}
        )
        mock_facets = {
            "schema": SchemaDatasetFacet(fields=[SchemaDatasetFacetFields(name="col", type="STRING")]),
            "documentation": DocumentationDatasetFacet(description="mock_description"),
        }
        mock_get_facets.return_value = mock_facets

        schema = "schema"
        table = "table"
        s3_bucket = "bucket"
        s3_key = "key"
        query = """
        SELECT
            c.customer_id,
            c.first_name,
            c.last_name,
            o.order_id,
            o.order_date,
            o.total_amount
        FROM
            schema1.customers c
        INNER JOIN
            schema2.orders o
        ON
            c.customer_id = o.customer_id
        ORDER BY
            o.order_date DESC;
        """
        op = RedshiftToS3Operator(
            schema=schema,  # should be ignored
            table=table,  # should be ignored
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            table_as_file_name=False,
            select_query=query,
            redshift_conn_id="redshift_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
        )
        op.execute(None)

        lineage = op.get_openlineage_facets_on_complete(None)

        assert len(lineage.inputs) == 2
        assert len(lineage.outputs) == 1
        assert lineage.outputs[0].name == s3_key
        assert lineage.outputs[0].namespace == f"s3://{s3_bucket}"
        assert lineage.inputs[0].name == "database.schema1.customers"
        assert lineage.inputs[0].namespace == "redshift://cluster.region:5439"
        assert lineage.inputs[1].name == "database.schema2.orders"
        assert lineage.inputs[1].namespace == "redshift://cluster.region:5439"

        assert lineage.inputs[0].facets == mock_facets
        assert lineage.outputs[0].facets == {}

    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.get_connection")
    @mock.patch("airflow.models.connection.Connection.get_connection_from_secrets")
    @mock.patch("boto3.session.Session")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_data.RedshiftDataHook.conn")
    @mock.patch(
        "airflow.providers.amazon.aws.hooks.redshift_data.RedshiftDataHook.region_name",
        new_callable=mock.PropertyMock,
    )
    @mock.patch("airflow.providers.amazon.aws.utils.openlineage.get_facets_from_redshift_table")
    def test_get_openlineage_facets_on_complete_using_redshift_data_api(
        self, mock_get_facets, mock_rs_region, mock_rs, mock_session, mock_connection, mock_hook
    ):
        """
        Using the Redshift Data API instead of the SQL-based connection
        """
        access_key = "aws_access_key_id"
        secret_key = "aws_secret_access_key"
        mock_session.return_value = Session(access_key, secret_key)
        mock_session.return_value.access_key = access_key
        mock_session.return_value.secret_key = secret_key
        mock_session.return_value.token = None

        mock_hook.return_value = Connection()
        mock_rs.execute_statement.return_value = {"Id": "STATEMENT_ID"}
        mock_rs.describe_statement.return_value = {"Status": "FINISHED"}

        mock_rs_region.return_value = "region"
        mock_facets = {
            "schema": SchemaDatasetFacet(fields=[SchemaDatasetFacetFields(name="col", type="STRING")]),
            "documentation": DocumentationDatasetFacet(description="mock_description"),
        }
        mock_get_facets.return_value = mock_facets

        schema = "schema"
        table = "table"
        s3_bucket = "bucket"
        s3_key = "key"

        # RS Data API params
        database = "database"
        cluster_identifier = "cluster"
        db_user = "db_user"
        secret_arn = "secret_arn"
        statement_name = "statement_name"

        op = RedshiftToS3Operator(
            schema=schema,
            table=table,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            redshift_conn_id="redshift_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            redshift_data_api_kwargs=dict(
                database=database,
                cluster_identifier=cluster_identifier,
                db_user=db_user,
                secret_arn=secret_arn,
                statement_name=statement_name,
            ),
        )
        op.execute(None)

        lineage = op.get_openlineage_facets_on_complete(None)

        assert len(lineage.inputs) == 1
        assert len(lineage.outputs) == 1
        assert lineage.outputs[0].name == f"{s3_key}/{table}_"
        assert lineage.outputs[0].namespace == f"s3://{s3_bucket}"
        assert lineage.inputs[0].name == f"database.{schema}.{table}"
        assert lineage.inputs[0].namespace == "redshift://cluster.region:5439"

        assert lineage.inputs[0].facets == mock_facets
        assert lineage.outputs[0].facets == {
            "columnLineage": ColumnLineageDatasetFacet(
                fields={
                    "col": Fields(
                        inputFields=[
                            InputField(
                                namespace="redshift://cluster.region:5439",
                                name=f"database.{schema}.{table}",
                                field="col",
                            )
                        ],
                        transformationType="IDENTITY",
                        transformationDescription="identical",
                    )
                }
            ),
            "schema": SchemaDatasetFacet(fields=[SchemaDatasetFacetFields(name="col", type="STRING")]),
        }

    @mock.patch("airflow.providers.amazon.aws.hooks.s3.S3Hook.get_connection")
    @mock.patch("airflow.models.connection.Connection.get_connection_from_secrets")
    @mock.patch("boto3.session.Session")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.RedshiftSQLHook.run")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_data.RedshiftDataHook.conn")
    @mock.patch(
        "airflow.providers.amazon.aws.hooks.redshift_data.RedshiftDataHook.region_name",
        new_callable=mock.PropertyMock,
    )
    @mock.patch("airflow.providers.amazon.aws.utils.openlineage.get_facets_from_redshift_table")
    def test_get_openlineage_facets_on_complete_data_and_sql_hooks_aligned(
        self, mock_get_facets, mock_rs_region, mock_rs, mock_run, mock_session, mock_connection, mock_hook
    ):
        """
        Ensuring both supported hooks - RedshiftDataHook and RedshiftSQLHook return same lineage.
        """
        access_key = "aws_access_key_id"
        secret_key = "aws_secret_access_key"
        mock_session.return_value = Session(access_key, secret_key)
        mock_session.return_value.access_key = access_key
        mock_session.return_value.secret_key = secret_key
        mock_session.return_value.token = None

        mock_connection.return_value = mock.MagicMock(
            schema="database", port=5439, host="cluster.id.region.redshift.amazonaws.com", extra_dejson={}
        )
        mock_hook.return_value = Connection()
        mock_rs.execute_statement.return_value = {"Id": "STATEMENT_ID"}
        mock_rs.describe_statement.return_value = {"Status": "FINISHED"}

        mock_rs_region.return_value = "region"
        mock_facets = {
            "schema": SchemaDatasetFacet(fields=[SchemaDatasetFacetFields(name="col", type="STRING")]),
            "documentation": DocumentationDatasetFacet(description="mock_description"),
        }
        mock_get_facets.return_value = mock_facets

        schema = "schema"
        table = "table"
        s3_bucket = "bucket"
        s3_key = "key"

        # RS Data API params
        database = "database"
        cluster_identifier = "cluster"
        db_user = "db_user"
        secret_arn = "secret_arn"
        statement_name = "statement_name"

        op_rs_data = RedshiftToS3Operator(
            schema=schema,
            table=table,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            redshift_conn_id="redshift_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            redshift_data_api_kwargs=dict(
                database=database,
                cluster_identifier=cluster_identifier,
                db_user=db_user,
                secret_arn=secret_arn,
                statement_name=statement_name,
            ),
        )
        op_rs_data.execute(None)
        rs_data_lineage = op_rs_data.get_openlineage_facets_on_complete(None)

        op_rs_sql = RedshiftToS3Operator(
            schema=schema,
            table=table,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            redshift_conn_id="redshift_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
        )
        op_rs_sql.execute(None)
        rs_sql_lineage = op_rs_sql.get_openlineage_facets_on_complete(None)

        assert len(rs_sql_lineage.inputs) == 1
        assert len(rs_sql_lineage.outputs) == 1
        assert rs_sql_lineage.inputs == rs_data_lineage.inputs
        assert rs_sql_lineage.outputs == rs_data_lineage.outputs
        assert rs_sql_lineage.job_facets == rs_data_lineage.job_facets
        assert rs_sql_lineage.run_facets == rs_data_lineage.run_facets



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow.decorators import task
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.amazon.aws.operators.s3 import (
    S3CreateBucketOperator,
    S3DeleteBucketOperator,
)
from airflow.providers.amazon.aws.transfers.gcs_to_s3 import GCSToS3Operator
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.cloud.operators.gcs import (
    GCSCreateBucketOperator,
    GCSDeleteBucketOperator,
)
from airflow.utils.trigger_rule import TriggerRule

from system.amazon.aws.utils import SystemTestContextBuilder

# Externally fetched variables:
GCP_PROJECT_ID = "GCP_PROJECT_ID"

sys_test_context_task = SystemTestContextBuilder().add_variable(GCP_PROJECT_ID).build()

DAG_ID = "example_gcs_to_s3"

with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["example"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]
    gcp_user_project = test_context[GCP_PROJECT_ID]

    s3_bucket = f"{env_id}-gcs-to-s3-bucket"
    s3_key = f"{env_id}-gcs-to-s3-key"

    create_s3_bucket = S3CreateBucketOperator(task_id="create_s3_bucket", bucket_name=s3_bucket)

    gcs_bucket = f"{env_id}-gcs-to-s3-bucket"
    gcs_key = f"{env_id}-gcs-to-s3-key"

    create_gcs_bucket = GCSCreateBucketOperator(
        task_id="create_gcs_bucket",
        bucket_name=gcs_bucket,
        resource={"billing": {"requesterPays": True}},
        project_id=gcp_user_project,
    )

    @task
    def upload_gcs_file(bucket_name: str, object_name: str, user_project: str):
        hook = GCSHook()
        with hook.provide_file_and_upload(
            bucket_name=bucket_name,
            object_name=object_name,
            user_project=user_project,
        ) as temp_file:
            temp_file.write(b"test")

    # [START howto_transfer_gcs_to_s3]
    gcs_to_s3 = GCSToS3Operator(
        task_id="gcs_to_s3",
        gcs_bucket=gcs_bucket,
        dest_s3_key=f"s3://{s3_bucket}/{s3_key}",
        replace=True,
        gcp_user_project=gcp_user_project,
    )
    # [END howto_transfer_gcs_to_s3]

    delete_s3_bucket = S3DeleteBucketOperator(
        task_id="delete_s3_bucket",
        bucket_name=s3_bucket,
        force_delete=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_gcs_bucket = GCSDeleteBucketOperator(
        task_id="delete_gcs_bucket",
        bucket_name=gcs_bucket,
        trigger_rule=TriggerRule.ALL_DONE,
        user_project=gcp_user_project,
    )

    chain(
        # TEST SETUP
        test_context,
        create_gcs_bucket,
        upload_gcs_file(gcs_bucket, gcs_key, gcp_user_project),
        create_s3_bucket,
        # TEST BODY
        gcs_to_s3,
        # TEST TEARDOWN
        delete_s3_bucket,
        delete_gcs_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)
